from __future__ import annotations

from typing import TYPE_CHECKING

from ...context import RunContext
from ..chat_prefs import ChatPrefsStore
from ..engine_overrides import (
    EngineOverrides,
    allowed_reasoning_levels,
    resolve_override_value,
)
from ..files import split_command_args
from ..topic_state import TopicStateStore
from ..topics import _topic_key
from ..types import TelegramIncomingMessage
from .overrides import (
    ENGINE_SOURCE_LABELS,
    OVERRIDE_SOURCE_LABELS,
    apply_engine_override,
    parse_set_args,
    require_admin_or_private,
    resolve_engine_selection,
)
from .reply import make_reply

if TYPE_CHECKING:
    from ..bridge import TelegramBridgeConfig

REASONING_USAGE = (
    "usage: `/reasoning`, `/reasoning set <level>`, "
    "`/reasoning set <engine> <level>`, or `/reasoning clear [engine]`"
)


async def _handle_reasoning_command(
    cfg: TelegramBridgeConfig,
    msg: TelegramIncomingMessage,
    args_text: str,
    ambient_context: RunContext | None,
    topic_store: TopicStateStore | None,
    chat_prefs: ChatPrefsStore | None,
    *,
    resolved_scope: str | None = None,
    scope_chat_ids: frozenset[int] | None = None,
) -> None:
    reply = make_reply(cfg, msg)
    tkey = (
        _topic_key(msg, cfg, scope_chat_ids=scope_chat_ids)
        if topic_store is not None
        else None
    )
    tokens = split_command_args(args_text)
    action = tokens[0].lower() if tokens else "show"
    engine_ids = {engine.lower() for engine in cfg.runtime.engine_ids}

    if action in {"show", ""}:
        selection = await resolve_engine_selection(
            cfg,
            msg,
            ambient_context=ambient_context,
            topic_store=topic_store,
            chat_prefs=chat_prefs,
            topic_key=tkey,
        )
        if selection is None:
            return
        engine, engine_source = selection
        topic_override = None
        if tkey is not None and topic_store is not None:
            topic_override = await topic_store.get_engine_override(
                tkey[0], tkey[1], engine
            )
        chat_override = None
        if chat_prefs is not None:
            chat_override = await chat_prefs.get_engine_override(msg.chat_id, engine)
        resolution = resolve_override_value(
            topic_override=topic_override,
            chat_override=chat_override,
            field="reasoning",
        )
        engine_line = f"engine: {engine} ({ENGINE_SOURCE_LABELS[engine_source]})"
        reasoning_value = resolution.value or "default"
        reasoning_line = (
            f"reasoning: {reasoning_value} "
            f"({OVERRIDE_SOURCE_LABELS[resolution.source]})"
        )
        topic_label = resolution.topic_value or "none"
        if tkey is None:
            topic_label = "none"
        chat_label = (
            "unavailable" if chat_prefs is None else resolution.chat_value or "none"
        )
        defaults_line = f"defaults: topic: {topic_label}, chat: {chat_label}"
        available_levels = ", ".join(allowed_reasoning_levels(engine))
        available_line = f"available levels: {available_levels}"
        await reply(
            text="\n\n".join(
                [engine_line, reasoning_line, defaults_line, available_line]
            )
        )
        return

    if action == "set":
        engine_arg, level = parse_set_args(tokens, engine_ids=engine_ids)
        if level is None:
            await reply(text=REASONING_USAGE)
            return
        if not await require_admin_or_private(
            cfg,
            msg,
            missing_sender="cannot verify sender for reasoning overrides.",
            failed_member="failed to verify reasoning override permissions.",
            denied="changing reasoning overrides is restricted to group admins.",
        ):
            return
        if engine_arg is None:
            selection = await resolve_engine_selection(
                cfg,
                msg,
                ambient_context=ambient_context,
                topic_store=topic_store,
                chat_prefs=chat_prefs,
                topic_key=tkey,
            )
            if selection is None:
                return
            engine, _ = selection
        else:
            engine = engine_arg
            if engine not in engine_ids:
                available = ", ".join(cfg.runtime.engine_ids)
                await reply(
                    text=f"unknown engine `{engine}`.\navailable engines: `{available}`"
                )
                return
        normalized_level = level.strip().lower()
        allowed = allowed_reasoning_levels(engine)
        if normalized_level not in allowed:
            await reply(
                text=(
                    f"unknown reasoning level `{level}`.\n"
                    f"available levels: {', '.join(allowed)}"
                )
            )
            return
        scope = await apply_engine_override(
            reply=reply,
            tkey=tkey,
            topic_store=topic_store,
            chat_prefs=chat_prefs,
            chat_id=msg.chat_id,
            engine=engine,
            update=lambda current: EngineOverrides(
                model=current.model if current is not None else None,
                reasoning=normalized_level,
            ),
            topic_unavailable="topic reasoning overrides are unavailable.",
            chat_unavailable="chat reasoning overrides are unavailable (no config path).",
        )
        if scope is None:
            return
        if scope == "topic":
            await reply(
                text=(
                    f"topic reasoning override set to `{normalized_level}` "
                    f"for `{engine}`.\n"
                    "If you want a clean start on the new setting, run `/new`."
                )
            )
            return
        await reply(
            text=(
                f"chat reasoning override set to `{normalized_level}` for `{engine}`.\n"
                "If you want a clean start on the new setting, run `/new`."
            )
        )
        return

    if action == "clear":
        engine = None
        if len(tokens) > 2:
            await reply(text=REASONING_USAGE)
            return
        if len(tokens) == 2:
            engine = tokens[1].strip().lower() or None
        if not await require_admin_or_private(
            cfg,
            msg,
            missing_sender="cannot verify sender for reasoning overrides.",
            failed_member="failed to verify reasoning override permissions.",
            denied="changing reasoning overrides is restricted to group admins.",
        ):
            return
        if engine is None:
            selection = await resolve_engine_selection(
                cfg,
                msg,
                ambient_context=ambient_context,
                topic_store=topic_store,
                chat_prefs=chat_prefs,
                topic_key=tkey,
            )
            if selection is None:
                return
            engine, _ = selection
        if engine not in engine_ids:
            available = ", ".join(cfg.runtime.engine_ids)
            await reply(
                text=f"unknown engine `{engine}`.\navailable engines: `{available}`"
            )
            return
        scope = await apply_engine_override(
            reply=reply,
            tkey=tkey,
            topic_store=topic_store,
            chat_prefs=chat_prefs,
            chat_id=msg.chat_id,
            engine=engine,
            update=lambda current: EngineOverrides(
                model=current.model if current is not None else None,
                reasoning=None,
            ),
            topic_unavailable="topic reasoning overrides are unavailable.",
            chat_unavailable="chat reasoning overrides are unavailable (no config path).",
        )
        if scope is None:
            return
        if scope == "topic":
            await reply(text="topic reasoning override cleared (using chat default).")
            return
        await reply(text="chat reasoning override cleared.")
        return

    await reply(text=REASONING_USAGE)



import re
from collections.abc import AsyncIterator
from typing import Any

import pytest

import yee88.runner as runner_module
from yee88.model import (
    ActionEvent,
    CompletedEvent,
    ResumeToken,
    StartedEvent,
    TakopiEvent,
)
from yee88.runner import (
    BaseRunner,
    JsonlRunState,
    JsonlSubprocessRunner,
    ResumeTokenMixin,
)


class _DummyRunner(ResumeTokenMixin, BaseRunner):
    engine = "dummy"
    resume_re = re.compile(r"(?im)^`?dummy resume (?P<token>[^`\s]+)`?$")

    async def run_impl(
        self, prompt: str, resume: ResumeToken | None
    ) -> AsyncIterator[StartedEvent | CompletedEvent]:
        token = resume or ResumeToken(engine=self.engine, value="token")
        yield StartedEvent(engine=self.engine, resume=token, title="dummy")
        yield CompletedEvent(
            engine=self.engine,
            ok=True,
            answer=prompt,
            resume=token,
        )


class _DummyJsonlRunner(JsonlSubprocessRunner):
    engine = "dummy-jsonl"

    def command(self) -> str:
        return "dummy"

    def build_args(
        self,
        prompt: str,
        resume: ResumeToken | None,
        *,
        state: object,
    ) -> list[str]:
        _ = prompt, resume, state
        return []

    def translate(
        self,
        data: Any,
        *,
        state: Any,
        resume: ResumeToken | None,
        found_session: ResumeToken | None,
    ) -> list[TakopiEvent]:
        _ = data, state, resume, found_session
        return []


class _BareJsonlRunner(JsonlSubprocessRunner):
    engine = "bare-jsonl"


class _RunJsonlRunner(_DummyJsonlRunner):
    def stdin_payload(
        self,
        prompt: str,
        resume: ResumeToken | None,
        *,
        state: Any,
    ) -> bytes | None:
        _ = prompt, resume, state
        return None

    async def iter_json_lines(self, stream: Any) -> AsyncIterator[bytes]:
        _ = stream
        yield b'{"type": "started", "resume": "sid"}'
        yield b'{"type": "completed", "resume": "sid"}'

    def translate(
        self,
        data: Any,
        *,
        state: Any,
        resume: ResumeToken | None,
        found_session: ResumeToken | None,
    ) -> list[TakopiEvent]:
        _ = state, resume, found_session
        token_value = "sid"
        if isinstance(data, dict) and isinstance(data.get("resume"), str):
            token_value = data["resume"]
        token = ResumeToken(engine=self.engine, value=token_value)
        if isinstance(data, dict) and data.get("type") == "started":
            return [StartedEvent(engine=self.engine, resume=token, title="t")]
        if isinstance(data, dict) and data.get("type") == "completed":
            return [
                CompletedEvent(engine=self.engine, ok=True, answer="done", resume=token)
            ]
        return []


class _BranchingJsonlRunner(_DummyJsonlRunner):
    def stdin_payload(
        self,
        prompt: str,
        resume: ResumeToken | None,
        *,
        state: Any,
    ) -> bytes | None:
        _ = prompt, resume, state
        return None

    async def iter_json_lines(self, stream: Any) -> AsyncIterator[bytes]:
        _ = stream
        yield b"raise"
        yield b""
        yield b"invalid"
        yield b'{"type": "translate_error"}'
        yield b'{"type": "started", "resume": "sid"}'
        yield b'{"type": "started", "resume": "sid"}'
        yield b'{"type": "completed", "resume": "sid"}'
        yield b'{"type": "after"}'

    def decode_jsonl(self, *, line: bytes) -> Any | None:
        if line == b"raise":
            raise ValueError("boom")
        if line == b"invalid":
            return None
        return super().decode_jsonl(line=line)

    def translate(
        self,
        data: Any,
        *,
        state: Any,
        resume: ResumeToken | None,
        found_session: ResumeToken | None,
    ) -> list[TakopiEvent]:
        _ = state, resume, found_session
        if isinstance(data, dict) and data.get("type") == "translate_error":
            raise RuntimeError("nope")
        token_value = "sid"
        if isinstance(data, dict) and isinstance(data.get("resume"), str):
            token_value = data["resume"]
        token = ResumeToken(engine=self.engine, value=token_value)
        if isinstance(data, dict) and data.get("type") == "started":
            return [StartedEvent(engine=self.engine, resume=token, title="t")]
        if isinstance(data, dict) and data.get("type") == "completed":
            return [
                CompletedEvent(engine=self.engine, ok=True, answer="done", resume=token)
            ]
        return []


@pytest.mark.anyio
async def test_base_runner_run_locked_handles_resume() -> None:
    runner = _DummyRunner()
    events = [evt async for evt in runner.run("hello", None)]
    assert isinstance(events[0], StartedEvent)
    assert isinstance(events[-1], CompletedEvent)

    resume = ResumeToken(engine=runner.engine, value="resume")
    resumed = [evt async for evt in runner.run("again", resume)]
    assert isinstance(resumed[0], StartedEvent)
    assert resumed[0].resume == resume


@pytest.mark.anyio
async def test_base_runner_rejects_wrong_resume_engine() -> None:
    runner = _DummyRunner()
    bad_resume = ResumeToken(engine="other", value="oops")
    with pytest.raises(RuntimeError):
        _ = [evt async for evt in runner.run("hello", bad_resume)]


@pytest.mark.anyio
async def test_base_runner_run_impl_not_implemented() -> None:
    class _BareRunner(BaseRunner):
        engine = "bare"

    runner = _BareRunner()
    with pytest.raises(NotImplementedError):
        _ = [evt async for evt in runner.run_impl("hello", None)]


def test_resume_token_format_and_extract() -> None:
    runner = _DummyRunner()
    token = ResumeToken(engine=runner.engine, value="abc")
    assert runner.format_resume(token) == "`dummy resume abc`"
    assert runner.is_resume_line("`dummy resume abc`") is True
    text = "`dummy resume first`\n`dummy resume second`"
    assert runner.extract_resume(text) == ResumeToken(
        engine=runner.engine, value="second"
    )
    assert runner.extract_resume(None) is None

    with pytest.raises(RuntimeError):
        runner.format_resume(ResumeToken(engine="other", value="bad"))


def test_session_lock_reuse() -> None:
    runner = _DummyRunner()
    token = ResumeToken(engine=runner.engine, value="one")
    lock1 = runner.lock_for(token)
    lock2 = runner.lock_for(token)
    other = runner.lock_for(ResumeToken(engine=runner.engine, value="two"))
    assert lock1 is lock2
    assert other is not lock1


@pytest.mark.anyio
async def test_run_with_resume_lock_passthrough() -> None:
    runner = _DummyRunner()
    events = [
        evt async for evt in runner.run_with_resume_lock("hello", None, runner.run_impl)
    ]
    assert events


def test_jsonl_helpers() -> None:
    runner = _DummyJsonlRunner()
    state = JsonlRunState()

    note1 = runner.next_note_id(state)
    note2 = runner.next_note_id(state)
    assert note1.endswith(".1")
    assert note2.endswith(".2")

    event = runner.note_event("warn", state=state)
    assert isinstance(event, ActionEvent)
    assert event.action.detail == {}

    invalid = runner.invalid_json_events(raw="x", line="{}", state=state)
    invalid_event = invalid[0]
    assert isinstance(invalid_event, ActionEvent)
    assert invalid_event.action.detail["line"] == "{}"

    assert runner.decode_jsonl(line=b'{"a": 1}') == {"a": 1}
    assert runner.decode_jsonl(line=b"{") is None

    err_events = runner.decode_error_events(
        raw="oops", line="{}", error=ValueError("nope"), state=state
    )
    err_event = err_events[0]
    assert isinstance(err_event, ActionEvent)
    assert err_event.action.detail["error"] == "nope"

    translated = runner.translate_error_events(
        data={"type": "foo", "item": {"type": "bar"}},
        error=ValueError("boom"),
        state=state,
    )
    translated_event = translated[0]
    assert isinstance(translated_event, ActionEvent)
    detail = translated_event.action.detail
    assert detail["type"] == "foo"
    assert detail["item_type"] == "bar"

    resume = ResumeToken(engine=runner.engine, value="sid")
    processed = runner.process_error_events(
        2, resume=resume, found_session=None, state=state
    )
    processed_event = processed[-1]
    assert isinstance(processed_event, CompletedEvent)
    assert processed_event.ok is False
    assert processed_event.resume == resume

    stream_end = runner.stream_end_events(
        resume=None, found_session=resume, state=state
    )
    stream_event = stream_end[-1]
    assert isinstance(stream_event, CompletedEvent)
    assert stream_event.resume == resume

    started = StartedEvent(engine=runner.engine, resume=resume, title="t")
    found, emit = runner.handle_started_event(
        started, expected_session=None, found_session=None
    )
    assert found == resume
    assert emit is True

    found, emit = runner.handle_started_event(
        started, expected_session=None, found_session=resume
    )
    assert found == resume
    assert emit is False

    mismatch = StartedEvent(engine="other", resume=resume, title="t")
    with pytest.raises(RuntimeError):
        runner.handle_started_event(mismatch, expected_session=None, found_session=None)

    other_resume = ResumeToken(engine=runner.engine, value="other")
    with pytest.raises(RuntimeError):
        runner.handle_started_event(
            StartedEvent(engine=runner.engine, resume=other_resume, title="t"),
            expected_session=resume,
            found_session=None,
        )

    with pytest.raises(RuntimeError):
        runner.handle_started_event(
            StartedEvent(engine=runner.engine, resume=other_resume, title="t"),
            expected_session=None,
            found_session=resume,
        )


def test_next_note_id_requires_state_field() -> None:
    runner = _DummyJsonlRunner()
    with pytest.raises(RuntimeError):
        runner.next_note_id(object())


def test_jsonl_base_methods_raise_and_defaults() -> None:
    runner = _BareJsonlRunner()
    with pytest.raises(NotImplementedError):
        runner.command()
    with pytest.raises(NotImplementedError):
        runner.build_args("hi", None, state=None)
    with pytest.raises(NotImplementedError):
        runner.translate(data={}, state=None, resume=None, found_session=None)
    assert runner.pipes_error_message().startswith("bare-jsonl")
    state = runner.new_state("hi", None)
    assert isinstance(state, JsonlRunState)
    assert runner.start_run("hi", None, state=state) is None
    assert runner.env(state=state) is None
    assert runner.stdin_payload("hi", None, state=state) == b"hi"


@pytest.mark.anyio
async def test_jsonl_run_impl_smoke(monkeypatch: pytest.MonkeyPatch) -> None:
    class _FakeProc:
        def __init__(self) -> None:
            self.stdout = object()
            self.stderr = object()
            self.stdin = None
            self.pid = 123

        async def wait(self) -> int:
            return 0

    class _FakeManager:
        def __init__(self, proc: _FakeProc) -> None:
            self._proc = proc

        async def __aenter__(self) -> _FakeProc:
            return self._proc

        async def __aexit__(self, exc_type, exc, tb) -> None:
            return None

    proc = _FakeProc()

    def fake_manage_subprocess(*args: Any, **kwargs: Any) -> _FakeManager:
        _ = args, kwargs
        return _FakeManager(proc)

    async def fake_drain_stderr(*args: Any, **kwargs: Any) -> None:
        _ = args, kwargs
        return None

    monkeypatch.setattr(runner_module, "manage_subprocess", fake_manage_subprocess)
    monkeypatch.setattr(runner_module, "drain_stderr", fake_drain_stderr)

    runner = _RunJsonlRunner()
    events = [evt async for evt in runner.run_impl("hello", None)]
    assert any(isinstance(evt, CompletedEvent) for evt in events)


@pytest.mark.anyio
async def test_jsonl_run_impl_branches(monkeypatch: pytest.MonkeyPatch) -> None:
    class _FakeProc:
        def __init__(self) -> None:
            self.stdout = object()
            self.stderr = object()
            self.stdin = None
            self.pid = 456

        async def wait(self) -> int:
            return 0

    class _FakeManager:
        def __init__(self, proc: _FakeProc) -> None:
            self._proc = proc

        async def __aenter__(self) -> _FakeProc:
            return self._proc

        async def __aexit__(self, exc_type, exc, tb) -> None:
            return None

    proc = _FakeProc()

    def fake_manage_subprocess(*args: Any, **kwargs: Any) -> _FakeManager:
        _ = args, kwargs
        return _FakeManager(proc)

    async def fake_drain_stderr(*args: Any, **kwargs: Any) -> None:
        _ = args, kwargs
        return None

    monkeypatch.setattr(runner_module, "manage_subprocess", fake_manage_subprocess)
    monkeypatch.setattr(runner_module, "drain_stderr", fake_drain_stderr)

    runner = _BranchingJsonlRunner()
    events = [evt async for evt in runner.run_impl("hello", None)]
    assert any(isinstance(evt, CompletedEvent) for evt in events)



from __future__ import annotations

import pytest

from yee88.runners import tool_actions
from yee88.utils.paths import reset_run_base_dir, set_run_base_dir


def test_tool_input_path_picks_first_match() -> None:
    tool_input = {"path": "src/main.py", "file": "ignored.txt"}
    assert (
        tool_actions.tool_input_path(tool_input, path_keys=("file", "path"))
        == "ignored.txt"
    )
    assert (
        tool_actions.tool_input_path(tool_input, path_keys=("path", "file"))
        == "src/main.py"
    )
    assert tool_actions.tool_input_path(tool_input, path_keys=("missing",)) is None


@pytest.mark.parametrize(
    ("tool_name", "tool_input", "expected_kind", "expected_title"),
    [
        ("bash", {"command": "echo hi"}, "command", "echo hi"),
        ("shell", {"command": "pwd"}, "command", "pwd"),
        ("edit", {"path": "src/app.py"}, "file_change", "src/app.py"),
        ("write", {"path": "notes.txt"}, "file_change", "notes.txt"),
        ("read", {"path": "README.md"}, "tool", "read: `README.md`"),
        ("glob", {"pattern": "*.py"}, "tool", "glob: `*.py`"),
        ("grep", {"pattern": "TODO"}, "tool", "grep: TODO"),
        ("find", {"pattern": "*.toml"}, "tool", "find: *.toml"),
        ("ls", {"path": "src"}, "tool", "ls: `src`"),
        ("websearch", {"query": "yee88"}, "web_search", "yee88"),
        (
            "webfetch",
            {"url": "https://example.com"},
            "web_search",
            "https://example.com",
        ),
        ("todowrite", {}, "note", "update todos"),
        ("todoread", {}, "note", "read todos"),
        ("askuserquestion", {}, "note", "ask user"),
        ("task", {"description": "do work"}, "subagent", "do work"),
        ("agent", {"prompt": "assist"}, "subagent", "assist"),
        ("unknown", {}, "tool", "unknown"),
    ],
)
def test_tool_kind_and_title_cases(
    tool_name: str,
    tool_input: dict[str, object],
    expected_kind: str,
    expected_title: str,
) -> None:
    token = set_run_base_dir(None)
    try:
        kind, title = tool_actions.tool_kind_and_title(
            tool_name,
            tool_input,
            path_keys=("path", "file"),
        )
    finally:
        reset_run_base_dir(token)

    assert kind == expected_kind
    assert title == expected_title


def test_tool_kind_and_title_task_kind_override() -> None:
    kind, title = tool_actions.tool_kind_and_title(
        "agent",
        {"description": "spawn worker"},
        path_keys=(),
        task_kind="warning",
    )

    assert kind == "warning"
    assert title == "spawn worker"



from __future__ import annotations

import re

ID_PATTERN = r"^[a-z0-9_]{1,32}$"
_ID_RE = re.compile(ID_PATTERN)

RESERVED_CLI_COMMANDS = frozenset({"config", "doctor", "init", "plugins"})
RESERVED_CHAT_COMMANDS = frozenset(
    {"cancel", "file", "new", "fork", "agent", "model", "reasoning", "trigger", "topic", "ctx"}
)
RESERVED_ENGINE_IDS = RESERVED_CLI_COMMANDS | RESERVED_CHAT_COMMANDS
RESERVED_COMMAND_IDS = RESERVED_CLI_COMMANDS | RESERVED_CHAT_COMMANDS


def is_valid_id(value: str) -> bool:
    return bool(_ID_RE.fullmatch(value))



from __future__ import annotations

from dataclasses import dataclass
from collections.abc import Callable

from .model import (
    Action,
    ActionEvent,
    ResumeToken,
    StartedEvent,
    TakopiEvent,
    TextDeltaEvent,
    TextFinishedEvent,
)


@dataclass(frozen=True, slots=True)
class ActionState:
    action: Action
    phase: str
    ok: bool | None
    display_phase: str
    completed: bool
    first_seen: int
    last_update: int


@dataclass(frozen=True, slots=True)
class ProgressState:
    engine: str
    action_count: int
    actions: tuple[ActionState, ...]
    resume: ResumeToken | None
    resume_line: str | None
    context_line: str | None
    model: str | None = None
    text_segments: tuple[str, ...] = ()
    streaming_text: str | None = None


class ProgressTracker:
    def __init__(self, *, engine: str) -> None:
        self.engine = engine
        self.resume: ResumeToken | None = None
        self.action_count = 0
        self._actions: dict[str, ActionState] = {}
        self._seq = 0
        self._text_segments: list[str] = []
        self._streaming_text: str | None = None

    def note_event(self, event: TakopiEvent) -> bool:
        match event:
            case StartedEvent(resume=resume):
                self.resume = resume
                return True
            case ActionEvent(action=action, phase=phase, ok=ok):
                if action.kind == "turn":
                    return False
                action_id = str(action.id or "")
                if not action_id:
                    return False
                completed = phase == "completed"
                existing = self._actions.get(action_id)
                has_open = existing is not None and not existing.completed
                is_update = phase == "updated" or (phase == "started" and has_open)
                display_phase = "updated" if is_update and not completed else phase

                self._seq += 1
                seq = self._seq

                if existing is None:
                    self.action_count += 1
                    first_seen = seq
                else:
                    first_seen = existing.first_seen
                self._actions[action_id] = ActionState(
                    action=action,
                    phase=phase,
                    ok=ok,
                    display_phase=display_phase,
                    completed=completed,
                    first_seen=first_seen,
                    last_update=seq,
                )
                return True
            case TextDeltaEvent(snapshot=snapshot):
                self._streaming_text = snapshot
                return True
            case TextFinishedEvent(text=text):
                if text:
                    self._text_segments.append(text)
                self._streaming_text = None
                return True
            case _:
                return False

    def set_resume(self, resume: ResumeToken | None) -> None:
        if resume is not None:
            self.resume = resume

    def snapshot(
        self,
        *,
        resume_formatter: Callable[[ResumeToken], str] | None = None,
        context_line: str | None = None,
        model: str | None = None,
    ) -> ProgressState:
        resume_line: str | None = None
        if self.resume is not None and resume_formatter is not None:
            resume_line = resume_formatter(self.resume)
        actions = tuple(
            sorted(self._actions.values(), key=lambda item: item.first_seen)
        )
        return ProgressState(
            engine=self.engine,
            action_count=self.action_count,
            actions=actions,
            resume=self.resume,
            resume_line=resume_line,
            context_line=context_line,
            model=model,
            text_segments=tuple(self._text_segments),
            streaming_text=self._streaming_text,
        )



from __future__ import annotations

import pytest

from yee88 import cli
from yee88.config import ConfigError
from yee88.lockfile import LockError
from yee88.settings import TakopiSettings


def _settings(overrides: dict | None = None) -> TakopiSettings:
    payload = {
        "transport": "telegram",
        "transports": {"telegram": {"bot_token": "token", "chat_id": 123}},
    }
    if overrides:
        payload.update(overrides)
    return TakopiSettings.model_validate(payload)


def test_parse_key_path_valid() -> None:
    assert cli._parse_key_path("transports.telegram.chat_id") == [
        "transports",
        "telegram",
        "chat_id",
    ]


def test_parse_key_path_invalid_segment() -> None:
    with pytest.raises(ConfigError):
        cli._parse_key_path("transports..chat_id")


def test_parse_value_toml_and_fallback() -> None:
    assert cli._parse_value("true") is True
    assert cli._parse_value("123") == 123
    assert cli._parse_value("not-toml") == "not-toml"


def test_toml_literal_and_error() -> None:
    assert cli._toml_literal("hello") == '"hello"'
    with pytest.raises(ConfigError):
        cli._toml_literal({"a": 1})


def test_flatten_config() -> None:
    flattened = cli._flatten_config(
        {"transports": {"telegram": {"chat_id": 123}}, "watch_config": True}
    )
    assert ("transports.telegram.chat_id", 123) in flattened
    assert ("watch_config", True) in flattened


def test_normalized_value_from_settings() -> None:
    settings = _settings()
    assert cli._normalized_value_from_settings(settings, ["transport"]) == "telegram"
    assert (
        cli._normalized_value_from_settings(
            settings, ["transports", "telegram", "chat_id"]
        )
        == 123
    )


def test_should_run_interactive(monkeypatch) -> None:
    class _Tty:
        def isatty(self) -> bool:
            return True

    class _NotTty:
        def isatty(self) -> bool:
            return False

    monkeypatch.setenv("TAKOPI_NO_INTERACTIVE", "1")
    assert cli._should_run_interactive() is False
    monkeypatch.delenv("TAKOPI_NO_INTERACTIVE")

    monkeypatch.setattr(cli.sys, "stdin", _Tty())
    monkeypatch.setattr(cli.sys, "stdout", _Tty())
    assert cli._should_run_interactive() is True

    monkeypatch.setattr(cli.sys, "stdin", _NotTty())
    monkeypatch.setattr(cli.sys, "stdout", _Tty())
    assert cli._should_run_interactive() is False


def test_resolve_transport_id_override(monkeypatch) -> None:
    assert cli._resolve_transport_id("  telegram ") == "telegram"
    with pytest.raises(ConfigError):
        cli._resolve_transport_id("   ")

    def _raise() -> None:
        raise ConfigError("boom")

    monkeypatch.setattr(cli, "load_or_init_config", _raise)
    assert cli._resolve_transport_id(None) == "telegram"


def test_doctor_file_checks() -> None:
    settings = _settings()
    checks = cli._doctor_file_checks(settings)
    assert checks[0].detail == "disabled"

    settings = _settings(
        {
            "transports": {
                "telegram": {
                    "bot_token": "token",
                    "chat_id": 1,
                    "files": {"enabled": True},
                }
            }
        }
    )
    checks = cli._doctor_file_checks(settings)
    assert checks[0].status == "warning"


def test_doctor_voice_checks(monkeypatch) -> None:
    settings = _settings()
    checks = cli._doctor_voice_checks(settings)
    assert checks[0].detail == "disabled"

    settings = _settings(
        {
            "transports": {
                "telegram": {
                    "bot_token": "token",
                    "chat_id": 1,
                    "voice_transcription": True,
                }
            }
        }
    )
    monkeypatch.delenv("OPENAI_API_KEY", raising=False)
    checks = cli._doctor_voice_checks(settings)
    assert checks[0].status == "error"
    assert checks[0].detail == "API key not set"

    settings_with_key = _settings(
        {
            "transports": {
                "telegram": {
                    "bot_token": "token",
                    "chat_id": 1,
                    "voice_transcription": True,
                    "voice_transcription_api_key": "local",
                }
            }
        }
    )
    checks = cli._doctor_voice_checks(settings_with_key)
    assert checks[0].status == "ok"
    assert checks[0].detail == "voice_transcription_api_key set"

    monkeypatch.setenv("OPENAI_API_KEY", "key")
    checks = cli._doctor_voice_checks(settings)
    assert checks[0].status == "ok"


def test_load_settings_optional(monkeypatch, tmp_path) -> None:
    def _raise() -> None:
        raise ConfigError("boom")

    monkeypatch.setattr(cli, "load_settings_if_exists", _raise)
    assert cli._load_settings_optional() == (None, None)

    monkeypatch.setattr(cli, "load_settings_if_exists", lambda: None)
    assert cli._load_settings_optional() == (None, None)

    settings = _settings()
    config_path = tmp_path / "yee88.toml"
    monkeypatch.setattr(cli, "load_settings_if_exists", lambda: (settings, config_path))
    assert cli._load_settings_optional() == (settings, config_path)


def test_acquire_config_lock_reports_error(monkeypatch, tmp_path) -> None:
    config_path = tmp_path / "yee88.toml"
    error = LockError(path=config_path, state="running")

    def _raise(*_args, **_kwargs):
        raise error

    messages: list[tuple[str, bool]] = []
    monkeypatch.setattr(cli, "acquire_lock", _raise)
    monkeypatch.setattr(
        cli.typer, "echo", lambda msg, err=False: messages.append((msg, err))
    )

    with pytest.raises(cli.typer.Exit) as exc:
        cli.acquire_config_lock(config_path, "token")

    assert exc.value.exit_code == 1
    assert any("already running" in msg for msg, _ in messages)



from yee88.model import ResumeToken
from yee88.runners.claude import ClaudeRunner
from yee88.runners.codex import CodexRunner
from yee88.runners.opencode import OpenCodeRunner, OpenCodeStreamState
from yee88.runners.pi import ENGINE as PI_ENGINE, PiRunner, PiStreamState
from yee88.runners.run_options import EngineRunOptions, apply_run_options


def test_codex_run_options_override_model_and_reasoning() -> None:
    runner = CodexRunner(codex_cmd="codex", extra_args=["-c", "notify=[]"])
    state = runner.new_state("hi", None)
    with apply_run_options(EngineRunOptions(model="gpt-4.1-mini", reasoning="low")):
        args = runner.build_args("hi", None, state=state)

    assert args == [
        "-c",
        "notify=[]",
        "--model",
        "gpt-4.1-mini",
        "-c",
        "model_reasoning_effort=low",
        "exec",
        "--json",
        "--skip-git-repo-check",
        "--color=never",
        "-",
    ]


def test_claude_run_options_override_model() -> None:
    runner = ClaudeRunner(claude_cmd="claude", model="claude-sonnet")
    with apply_run_options(EngineRunOptions(model="claude-opus")):
        args = runner.build_args("hi", None, state=None)

    assert "--model" in args
    model_idx = args.index("--model") + 1
    assert args[model_idx] == "claude-opus"


def test_opencode_run_options_override_model() -> None:
    runner = OpenCodeRunner(opencode_cmd="opencode", model="claude-sonnet")
    state = OpenCodeStreamState()
    with apply_run_options(EngineRunOptions(model="gpt-4o-mini")):
        args = runner.build_args("hi", None, state=state)

    assert "--model" in args
    model_idx = args.index("--model") + 1
    assert args[model_idx] == "gpt-4o-mini"


def test_pi_run_options_override_model() -> None:
    runner = PiRunner(extra_args=[], model="pi-default", provider=None)
    state = PiStreamState(resume=ResumeToken(engine=PI_ENGINE, value="sess.jsonl"))
    with apply_run_options(EngineRunOptions(model="pi-override")):
        args = runner.build_args("hi", None, state=state)

    assert "--model" in args
    model_idx = args.index("--model") + 1
    assert args[model_idx] == "pi-override"



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Protocol

type ChannelId = int | str
type MessageId = int | str
type ThreadId = int | str


@dataclass(frozen=True, slots=True)
class MessageRef:
    channel_id: ChannelId
    message_id: MessageId
    raw: Any | None = field(default=None, compare=False, hash=False)
    thread_id: ThreadId | None = field(default=None, compare=False, hash=False)
    sender_id: int | None = field(default=None, compare=False, hash=False)


@dataclass(frozen=True, slots=True)
class RenderedMessage:
    text: str
    extra: dict[str, Any] = field(default_factory=dict)


@dataclass(frozen=True, slots=True)
class SendOptions:
    reply_to: MessageRef | None = None
    notify: bool = True
    replace: MessageRef | None = None
    thread_id: ThreadId | None = None


class Transport(Protocol):
    async def close(self) -> None: ...

    async def send(
        self,
        *,
        channel_id: ChannelId,
        message: RenderedMessage,
        options: SendOptions | None = None,
    ) -> MessageRef | None: ...

    async def edit(
        self,
        *,
        ref: MessageRef,
        message: RenderedMessage,
        wait: bool = True,
    ) -> MessageRef | None: ...

    async def delete(self, *, ref: MessageRef) -> bool: ...



# /// script
# requires-python = ">=3.14"
# dependencies = [
#     "mistune>=3.2.0",
#     "requests>=2.32.5",
#     "sulguk>=0.11.1",
# ]
# ///
import mistune
import sulguk
import requests
import os
import re

repo = os.environ["REPO"]
tag = os.environ["TAG_NAME"]
bot_token = os.environ["TELEGRAM_BOT_TOKEN"]
chat_id = os.environ["TELEGRAM_CHAT_ID"]

PULL_RE = re.compile(rf"(https://github.com/{repo}/pull/(\d+))")
COMPARE_RE = re.compile(rf"(https://github.com/{repo}/compare/(.*))")

resp = requests.get(f"https://api.github.com/repos/{repo}/releases/tags/{tag}")
body = resp.json()["body"]
lines = body.splitlines()

header = f"release **{repo} {tag}**"
pulls = "\n".join(
    [PULL_RE.sub(r"[#\2](\1)", line) for line in lines if PULL_RE.search(line)]
)
compare = COMPARE_RE.sub(r"compare [\2](\1)", COMPARE_RE.search(body).group(0))

message = "\n\n".join([header, pulls, compare])
print(message)

html = mistune.html(message)
rendered = sulguk.transform_html(html)

payload = {
    "chat_id": chat_id,
    "text": rendered.text,
    "entities": rendered.entities,
    "link_preview_options": {"is_disabled": True},
}
resp = requests.post(
    f"https://api.telegram.org/bot{bot_token}/sendMessage", json=payload
)
resp.raise_for_status()
print(f"\nsent to {chat_id}")



from pathlib import Path
import tomllib

from typer.testing import CliRunner

from yee88 import cli


def _write_min_config(path: Path) -> None:
    path.write_text(
        'transport = "telegram"\n'
        "\n"
        "[transports.telegram]\n"
        'bot_token = "token"\n'
        "chat_id = 123\n",
        encoding="utf-8",
    )


def test_config_list_outputs_flattened(tmp_path: Path) -> None:
    config_path = tmp_path / "yee88.toml"
    config_path.write_text(
        'transport = "telegram"\n'
        "watch_config = true\n"
        "\n"
        "[transports.telegram]\n"
        'bot_token = "token"\n'
        "chat_id = 123\n",
        encoding="utf-8",
    )

    runner = CliRunner()
    result = runner.invoke(
        cli.create_app(),
        ["config", "list", "--config-path", str(config_path)],
    )

    assert result.exit_code == 0
    lines = [line.strip() for line in result.output.splitlines() if line.strip()]
    assert 'transport = "telegram"' in lines
    assert "watch_config = true" in lines
    assert 'transports.telegram.bot_token = "token"' in lines
    assert "transports.telegram.chat_id = 123" in lines
    assert not any(line.startswith("default_engine") for line in lines)


def test_config_get_outputs_literal_and_table_error(tmp_path: Path) -> None:
    config_path = tmp_path / "yee88.toml"
    _write_min_config(config_path)

    runner = CliRunner()
    result = runner.invoke(
        cli.create_app(),
        [
            "config",
            "get",
            "transports.telegram.chat_id",
            "--config-path",
            str(config_path),
        ],
    )

    assert result.exit_code == 0
    assert result.output.strip() == "123"

    result = runner.invoke(
        cli.create_app(),
        [
            "config",
            "get",
            "transports.telegram",
            "--config-path",
            str(config_path),
        ],
    )

    assert result.exit_code == 2
    assert "table" in result.output


def test_config_get_missing_key(tmp_path: Path) -> None:
    config_path = tmp_path / "yee88.toml"
    _write_min_config(config_path)

    runner = CliRunner()
    result = runner.invoke(
        cli.create_app(),
        ["config", "get", "nope", "--config-path", str(config_path)],
    )

    assert result.exit_code == 1
    assert result.output == ""


def test_config_set_parses_and_writes(tmp_path: Path) -> None:
    config_path = tmp_path / "yee88.toml"
    _write_min_config(config_path)

    runner = CliRunner()
    result = runner.invoke(
        cli.create_app(),
        ["config", "set", "watch_config", "true", "--config-path", str(config_path)],
    )
    assert result.exit_code == 0

    result = runner.invoke(
        cli.create_app(),
        [
            "config",
            "set",
            "default_engine",
            "openai",
            "--config-path",
            str(config_path),
        ],
    )
    assert result.exit_code == 0

    result = runner.invoke(
        cli.create_app(),
        [
            "config",
            "set",
            "watch_config",
            "False",
            "--config-path",
            str(config_path),
        ],
    )
    assert result.exit_code == 0

    result = runner.invoke(
        cli.create_app(),
        [
            "config",
            "set",
            "transports.telegram.chat_id",
            "456",
            "--config-path",
            str(config_path),
        ],
    )
    assert result.exit_code == 0

    data = tomllib.loads(config_path.read_text(encoding="utf-8"))
    assert data["watch_config"] is False
    assert data["default_engine"] == "openai"
    assert data["transports"]["telegram"]["chat_id"] == 456


def test_config_unset_prunes_tables(tmp_path: Path) -> None:
    config_path = tmp_path / "yee88.toml"
    config_path.write_text(
        'transport = "telegram"\n'
        "\n"
        "[transports.telegram]\n"
        'bot_token = "token"\n'
        "chat_id = 123\n"
        "\n"
        "[projects.foo]\n"
        'path = "/tmp/repo"\n',
        encoding="utf-8",
    )

    runner = CliRunner()
    result = runner.invoke(
        cli.create_app(),
        ["config", "unset", "projects.foo", "--config-path", str(config_path)],
    )

    assert result.exit_code == 0
    data = tomllib.loads(config_path.read_text(encoding="utf-8"))
    assert "projects" not in data


def test_config_set_schema_validation_error(tmp_path: Path) -> None:
    config_path = tmp_path / "yee88.toml"
    config_path.write_text(
        'transport = "telegram"\n'
        "\n"
        "[transports.telegram]\n"
        'bot_token = "token"\n'
        "chat_id = 123\n"
        "\n"
        "[projects.foo]\n"
        'path = "/tmp/repo"\n',
        encoding="utf-8",
    )

    runner = CliRunner()
    result = runner.invoke(
        cli.create_app(),
        [
            "config",
            "set",
            "projects.foo.extra",
            "nope",
            "--config-path",
            str(config_path),
        ],
    )

    assert result.exit_code == 2
    data = tomllib.loads(config_path.read_text(encoding="utf-8"))
    assert "extra" not in data.get("projects", {}).get("foo", {})



from dataclasses import dataclass
from typing import Optional


@dataclass
class CronJob:
    id: str
    schedule: str
    message: str
    project: str
    enabled: bool = True
    last_run: str = ""
    next_run: str = ""
    one_time: bool = False
    engine: Optional[str] = None
    model: Optional[str] = None



from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING

import anyio

from ...commands import CommandContext, get_command
from ...config import ConfigError
from ...logging import get_logger
from ...model import EngineId, ResumeToken
from ...runners.run_options import EngineRunOptions
from ...runner_bridge import RunningTasks
from ...scheduler import ThreadScheduler
from ...transport import MessageRef
from ..files import split_command_args
from ..types import TelegramIncomingMessage
from .executor import _TelegramCommandExecutor

if TYPE_CHECKING:
    from ..bridge import TelegramBridgeConfig

logger = get_logger(__name__)


async def _dispatch_command(
    cfg: TelegramBridgeConfig,
    msg: TelegramIncomingMessage,
    text: str,
    command_id: str,
    args_text: str,
    running_tasks: RunningTasks,
    scheduler: ThreadScheduler,
    on_thread_known: Callable[[ResumeToken, anyio.Event], Awaitable[None]] | None,
    stateful_mode: bool,
    default_engine_override: EngineId | None,
    engine_overrides_resolver: Callable[[EngineId], Awaitable[EngineRunOptions | None]]
    | None,
) -> None:
    allowlist = cfg.runtime.allowlist
    chat_id = msg.chat_id
    user_msg_id = msg.message_id
    reply_ref = (
        MessageRef(
            channel_id=chat_id,
            message_id=msg.reply_to_message_id,
            thread_id=msg.thread_id,
        )
        if msg.reply_to_message_id is not None
        else None
    )
    executor = _TelegramCommandExecutor(
        exec_cfg=cfg.exec_cfg,
        runtime=cfg.runtime,
        running_tasks=running_tasks,
        scheduler=scheduler,
        on_thread_known=on_thread_known,
        engine_overrides_resolver=engine_overrides_resolver,
        chat_id=chat_id,
        user_msg_id=user_msg_id,
        thread_id=msg.thread_id,
        show_resume_line=cfg.show_resume_line,
        stateful_mode=stateful_mode,
        default_engine_override=default_engine_override,
    )
    message_ref = MessageRef(
        channel_id=chat_id,
        message_id=user_msg_id,
        thread_id=msg.thread_id,
        sender_id=msg.sender_id,
        raw=msg.raw,
    )
    try:
        backend = get_command(command_id, allowlist=allowlist, required=False)
    except ConfigError as exc:
        await executor.send(f"error:\n{exc}", reply_to=message_ref, notify=True)
        return
    if backend is None:
        return
    try:
        plugin_config = cfg.runtime.plugin_config(command_id)
    except ConfigError as exc:
        await executor.send(f"error:\n{exc}", reply_to=message_ref, notify=True)
        return
    ctx = CommandContext(
        command=command_id,
        text=text,
        args_text=args_text,
        args=split_command_args(args_text),
        message=message_ref,
        reply_to=reply_ref,
        reply_text=msg.reply_to_text,
        config_path=cfg.runtime.config_path,
        plugin_config=plugin_config,
        runtime=cfg.runtime,
        executor=executor,
    )
    try:
        result = await backend.handle(ctx)
    except Exception as exc:
        logger.exception(
            "command.failed",
            command=command_id,
            error=str(exc),
            error_type=exc.__class__.__name__,
        )
        await executor.send(f"error:\n{exc}", reply_to=message_ref, notify=True)
        return
    if result is not None:
        reply_to = message_ref if result.reply_to is None else result.reply_to
        await executor.send(result.text, reply_to=reply_to, notify=result.notify)



from __future__ import annotations

from pathlib import Path

import pytest

from yee88.config import ConfigError, read_config
from yee88.settings import (
    TakopiSettings,
    load_settings,
    load_settings_if_exists,
    require_telegram,
    validate_settings_data,
)


def test_load_settings_from_toml(tmp_path: Path) -> None:
    config_path = tmp_path / "yee88.toml"
    config_path.write_text(
        'transport = "telegram"\n\n'
        "[transports.telegram]\n"
        'bot_token = "token"\n'
        "chat_id = 123\n\n"
        "[codex]\n"
        'model = "gpt-4"\n',
        encoding="utf-8",
    )

    settings, loaded_path = load_settings(config_path)

    assert loaded_path == config_path
    assert settings.transport == "telegram"
    assert settings.transports.telegram.chat_id == 123
    assert settings.engine_config("codex", config_path=config_path)["model"] == "gpt-4"

    token, chat_id = require_telegram(settings, config_path)
    assert token == "token"
    assert chat_id == 123

    assert settings.transports.telegram.bot_token == "token"


def test_env_overrides_toml(tmp_path: Path, monkeypatch) -> None:
    config_path = tmp_path / "yee88.toml"
    config_path.write_text(
        'default_engine = "codex"\n'
        'transport = "telegram"\n\n'
        "[transports.telegram]\n"
        'bot_token = "token"\n'
        "chat_id = 123\n",
        encoding="utf-8",
    )
    monkeypatch.setenv("TAKOPI__DEFAULT_ENGINE", "claude")

    settings, _ = load_settings(config_path)

    assert settings.default_engine == "claude"


def test_legacy_keys_migrated(tmp_path: Path) -> None:
    config_path = tmp_path / "yee88.toml"
    config_path.write_text('bot_token = "token"\nchat_id = 123\n', encoding="utf-8")

    settings, loaded_path = load_settings(config_path)

    assert loaded_path == config_path
    assert settings.transports.telegram.chat_id == 123
    raw = read_config(config_path)
    assert "bot_token" not in raw
    assert "chat_id" not in raw
    assert raw["transports"]["telegram"]["bot_token"] == "token"
    assert raw["transports"]["telegram"]["chat_id"] == 123
    assert raw["transport"] == "telegram"


def test_validate_settings_data_rejects_invalid_bot_token_type(tmp_path: Path) -> None:
    config_path = tmp_path / "yee88.toml"
    data = {
        "transport": "telegram",
        "transports": {"telegram": {"bot_token": 123, "chat_id": 123}},
    }

    with pytest.raises(ConfigError, match="bot_token"):
        validate_settings_data(data, config_path=config_path)


def test_validate_settings_data_rejects_empty_default_engine(tmp_path: Path) -> None:
    config_path = tmp_path / "yee88.toml"
    data = {
        "default_engine": "   ",
        "transport": "telegram",
        "transports": {"telegram": {"bot_token": "token", "chat_id": 123}},
    }

    with pytest.raises(ConfigError, match="default_engine"):
        validate_settings_data(data, config_path=config_path)


def test_validate_settings_data_rejects_empty_default_project(tmp_path: Path) -> None:
    config_path = tmp_path / "yee88.toml"
    data = {
        "default_project": "   ",
        "transports": {"telegram": {"bot_token": "token", "chat_id": 123}},
    }

    with pytest.raises(ConfigError, match="default_project"):
        validate_settings_data(data, config_path=config_path)


def test_validate_settings_data_rejects_empty_project_path(tmp_path: Path) -> None:
    config_path = tmp_path / "yee88.toml"
    data = {
        "projects": {"z80": {"path": "   "}},
        "transports": {"telegram": {"bot_token": "token", "chat_id": 123}},
    }

    with pytest.raises(ConfigError, match="path"):
        validate_settings_data(data, config_path=config_path)


def test_engine_config_none_and_invalid(tmp_path: Path) -> None:
    config_path = tmp_path / "yee88.toml"
    settings = TakopiSettings.model_validate(
        {
            "transport": "telegram",
            "transports": {"telegram": {"bot_token": "token", "chat_id": 123}},
            "codex": None,
        }
    )
    assert settings.engine_config("codex", config_path=config_path) == {}

    settings = TakopiSettings.model_validate(
        {
            "transport": "telegram",
            "transports": {"telegram": {"bot_token": "token", "chat_id": 123}},
            "codex": "nope",
        }
    )
    with pytest.raises(ConfigError, match="codex"):
        settings.engine_config("codex", config_path=config_path)


def test_transport_config_telegram_and_extra(tmp_path: Path) -> None:
    config_path = tmp_path / "yee88.toml"
    settings = TakopiSettings.model_validate(
        {
            "transport": "telegram",
            "transports": {"telegram": {"bot_token": "token", "chat_id": 123}},
        }
    )
    telegram = settings.transport_config("telegram", config_path=config_path)
    assert telegram["bot_token"] == "token"
    assert telegram["chat_id"] == 123

    settings = TakopiSettings.model_validate(
        {
            "transport": "telegram",
            "transports": {
                "telegram": {"bot_token": "token", "chat_id": 123},
                "discord": None,
            },
        }
    )
    assert settings.transport_config("discord", config_path=config_path) == {}

    settings = TakopiSettings.model_validate(
        {
            "transport": "telegram",
            "transports": {
                "telegram": {"bot_token": "token", "chat_id": 123},
                "discord": "nope",
            },
        }
    )
    with pytest.raises(ConfigError, match=r"transports\.discord"):
        settings.transport_config("discord", config_path=config_path)


def test_bot_token_none_rejected(tmp_path: Path) -> None:
    config_path = tmp_path / "yee88.toml"
    data = {
        "transport": "telegram",
        "transports": {"telegram": {"bot_token": None, "chat_id": 123}},
    }
    with pytest.raises(ConfigError, match="bot_token"):
        validate_settings_data(data, config_path=config_path)


def test_require_telegram_rejects_non_telegram_transport(tmp_path: Path) -> None:
    config_path = tmp_path / "yee88.toml"
    settings = TakopiSettings.model_validate(
        {
            "transport": "discord",
            "transports": {"telegram": {"bot_token": "token", "chat_id": 123}},
        }
    )
    with pytest.raises(ConfigError, match="Unsupported transport"):
        require_telegram(settings, config_path)


def test_load_settings_if_exists_missing(tmp_path: Path) -> None:
    config_path = tmp_path / "missing.toml"
    assert load_settings_if_exists(config_path) is None


def test_load_settings_missing_file(tmp_path: Path) -> None:
    config_path = tmp_path / "missing.toml"
    with pytest.raises(ConfigError, match="Missing config file"):
        load_settings(config_path)


def test_load_settings_if_exists_loads(tmp_path: Path) -> None:
    config_path = tmp_path / "yee88.toml"
    config_path.write_text(
        'transport = "telegram"\n\n[transports.telegram]\n'
        'bot_token = "token"\nchat_id = 123\n',
        encoding="utf-8",
    )

    loaded = load_settings_if_exists(config_path)
    assert loaded is not None
    settings, loaded_path = loaded
    assert loaded_path == config_path


def test_load_settings_if_exists_rejects_non_file(tmp_path: Path) -> None:
    config_path = tmp_path / "config_dir"
    config_path.mkdir()
    with pytest.raises(ConfigError, match="exists but is not a file"):
        load_settings_if_exists(config_path)


def test_load_settings_rejects_non_file(tmp_path: Path) -> None:
    config_path = tmp_path / "config_dir"
    config_path.mkdir()
    with pytest.raises(ConfigError, match="exists but is not a file"):
        load_settings(config_path)



from __future__ import annotations

from collections.abc import AsyncIterator
import sys
from typing import Any

import anyio
from anyio.abc import ByteReceiveStream
from anyio.streams.buffered import BufferedByteReceiveStream

from ..logging import log_pipeline


async def iter_bytes_lines(stream: ByteReceiveStream) -> AsyncIterator[bytes]:
    buffered = BufferedByteReceiveStream(stream)
    while True:
        try:
            line = await buffered.receive_until(b"\n", sys.maxsize)
        except anyio.IncompleteRead:
            return
        yield line


async def drain_stderr(
    stream: ByteReceiveStream,
    logger: Any,
    tag: str,
) -> None:
    try:
        async for line in iter_bytes_lines(stream):
            text = line.decode("utf-8", errors="replace")
            log_pipeline(
                logger,
                "subprocess.stderr",
                tag=tag,
                line=text,
            )
    except Exception as exc:  # noqa: BLE001
        log_pipeline(
            logger,
            "subprocess.stderr.error",
            tag=tag,
            error=str(exc),
        )



from dataclasses import replace

from yee88.settings import TelegramTopicsSettings
from yee88.telegram.topics import _resolve_topics_scope_raw, _topics_command_error
from tests.telegram_fakes import FakeTransport, make_cfg


def test_resolve_topics_scope_raw() -> None:
    resolved, chat_ids = _resolve_topics_scope_raw("auto", 1, ())
    assert resolved == "main"
    assert chat_ids == frozenset({1})

    resolved, chat_ids = _resolve_topics_scope_raw("projects", 1, (2, 3))
    assert resolved == "projects"
    assert chat_ids == frozenset({2, 3})

    resolved, chat_ids = _resolve_topics_scope_raw("all", 1, (2,))
    assert resolved == "all"
    assert chat_ids == frozenset({1, 2})


def test_topics_command_error_for_wrong_chat() -> None:
    transport = FakeTransport()
    cfg = replace(
        make_cfg(transport),
        topics=TelegramTopicsSettings(enabled=True, scope="main"),
    )
    error = _topics_command_error(
        cfg,
        chat_id=999,
        resolved_scope="main",
        scope_chat_ids=frozenset({cfg.chat_id}),
    )
    assert error == "topics commands are only available in the main chat."



from __future__ import annotations

import re
import uuid
from collections.abc import AsyncIterator, Awaitable, Callable, Iterable
from dataclasses import dataclass, replace

import anyio

from ..model import (
    ActionEvent,
    CompletedEvent,
    EngineId,
    ResumeToken,
    StartedEvent,
    TakopiEvent,
)
from ..runner import ResumeTokenMixin, Runner, SessionLockMixin

ENGINE: EngineId = "mock"


@dataclass(frozen=True, slots=True)
class Emit:
    event: TakopiEvent
    at: float | None = None


@dataclass(frozen=True, slots=True)
class Advance:
    now: float


@dataclass(frozen=True, slots=True)
class Sleep:
    seconds: float


@dataclass(frozen=True, slots=True)
class Wait:
    event: anyio.Event


@dataclass(frozen=True, slots=True)
class Return:
    answer: str


@dataclass(frozen=True, slots=True)
class Raise:
    error: Exception


type ScriptStep = Emit | Advance | Sleep | Wait | Return | Raise


def _resume_token(engine: EngineId, value: str | None) -> ResumeToken:
    return ResumeToken(engine=engine, value=value or uuid.uuid4().hex)


class MockRunner(SessionLockMixin, ResumeTokenMixin, Runner):
    engine: EngineId
    model: str | None

    def __init__(
        self,
        *,
        events: Iterable[TakopiEvent] | None = None,
        answer: str = "",
        engine: EngineId = ENGINE,
        resume_value: str | None = None,
        title: str | None = None,
        model: str | None = None,
    ) -> None:
        self.engine = engine
        self.model = model
        self._events = list(events or [])
        self._answer = answer
        self._resume_value = resume_value
        self.title = title or str(engine).title()
        engine_name = re.escape(str(engine))
        self.resume_re = re.compile(
            rf"(?im)^\s*`?{engine_name}\s+resume\s+(?P<token>[^`\s]+)`?\s*$"
        )

    async def run(
        self, prompt: str, resume: ResumeToken | None
    ) -> AsyncIterator[TakopiEvent]:
        token_value = None
        if resume is not None:
            if resume.engine != self.engine:
                raise RuntimeError(
                    f"resume token is for engine {resume.engine!r}, not {self.engine!r}"
                )
            token_value = resume.value
        if token_value is None:
            token_value = self._resume_value
        token = _resume_token(self.engine, token_value)
        session_evt = StartedEvent(
            engine=self.engine,
            resume=token,
            title=self.title,
        )
        lock = self.lock_for(token)
        async with lock:
            yield session_evt

            for event in self._events:
                event_out: TakopiEvent = event
                if (
                    isinstance(event_out, ActionEvent)
                    and event_out.phase == "completed"
                    and event_out.ok is None
                ):
                    event_out = replace(event_out, ok=True)
                yield event_out
                await anyio.sleep(0)

            yield CompletedEvent(
                engine=self.engine,
                resume=token,
                ok=True,
                answer=self._answer,
            )


class ScriptRunner(MockRunner):
    def __init__(
        self,
        script: Iterable[ScriptStep],
        *,
        engine: EngineId = ENGINE,
        resume_value: str | None = None,
        emit_session_start: bool = True,
        sleep: Callable[[float], Awaitable[None]] = anyio.sleep,
        advance: Callable[[float], None] | None = None,
        default_answer: str = "",
        title: str | None = None,
    ) -> None:
        super().__init__(
            events=[],
            answer=default_answer,
            engine=engine,
            resume_value=resume_value,
            title=title,
        )
        self.calls: list[tuple[str, ResumeToken | None]] = []
        self._script = list(script)
        self._emit_session_start = emit_session_start
        self._sleep = sleep
        self._advance = advance

    def _advance_to(self, now: float) -> None:
        if self._advance is None:
            raise RuntimeError("ScriptRunner advance callback is not configured.")
        self._advance(now)

    async def run(
        self, prompt: str, resume: ResumeToken | None
    ) -> AsyncIterator[TakopiEvent]:
        self.calls.append((prompt, resume))
        token_value = None
        if resume is not None:
            if resume.engine != self.engine:
                raise RuntimeError(
                    f"resume token is for engine {resume.engine!r}, not {self.engine!r}"
                )
            token_value = resume.value
        if token_value is None:
            token_value = self._resume_value
        token = _resume_token(self.engine, token_value)
        session_evt = StartedEvent(
            engine=self.engine,
            resume=token,
            title=self.title,
        )
        lock = self.lock_for(token)

        async with lock:
            if self._emit_session_start:
                yield session_evt
                await anyio.sleep(0)

            for step in self._script:
                if isinstance(step, Emit):
                    if step.at is not None:
                        self._advance_to(step.at)
                    event_out: TakopiEvent = step.event
                    if (
                        isinstance(event_out, ActionEvent)
                        and event_out.phase == "completed"
                        and event_out.ok is None
                    ):
                        event_out = replace(event_out, ok=True)
                    yield event_out
                    await anyio.sleep(0)
                    continue
                if isinstance(step, Advance):
                    self._advance_to(step.now)
                    continue
                if isinstance(step, Sleep):
                    await self._sleep(step.seconds)
                    continue
                if isinstance(step, Wait):
                    await step.event.wait()
                    continue
                if isinstance(step, Raise):
                    raise step.error
                if isinstance(step, Return):
                    yield CompletedEvent(
                        engine=self.engine,
                        resume=token,
                        ok=True,
                        answer=step.answer,
                    )
                    return
                raise RuntimeError(f"Unhandled script step: {step!r}")

            yield CompletedEvent(
                engine=self.engine,
                resume=token,
                ok=True,
                answer=self._answer,
            )



from __future__ import annotations

# ruff: noqa: F401

from collections.abc import Callable
import sys
from pathlib import Path

import typer

from .. import __version__
from ..config import (
    ConfigError,
    HOME_CONFIG_PATH,
    load_or_init_config,
    write_config,
)
from ..config_migrations import migrate_config
from ..commands import get_command
from ..engines import get_backend, list_backend_ids
from ..ids import RESERVED_CHAT_COMMANDS, RESERVED_COMMAND_IDS, RESERVED_ENGINE_IDS
from ..lockfile import LockError, LockHandle, acquire_lock, token_fingerprint
from ..logging import setup_logging
from ..runtime_loader import build_runtime_spec, resolve_plugins_allowlist
from ..settings import (
    TakopiSettings,
    load_settings,
    load_settings_if_exists,
    validate_settings_data,
)
from ..plugins import (
    COMMAND_GROUP,
    ENGINE_GROUP,
    TRANSPORT_GROUP,
    entrypoint_distribution_name,
    get_load_errors,
    is_entrypoint_allowed,
    list_entrypoints,
    normalize_allowlist,
)
from ..transports import get_transport
from ..utils.git import resolve_default_base, resolve_main_worktree_root
from ..telegram import onboarding
from ..telegram.client import TelegramClient
from ..telegram.topics import _validate_topics_setup_for
from .doctor import (
    DoctorCheck,
    DoctorStatus,
    _doctor_file_checks,
    _doctor_telegram_checks,
    _doctor_voice_checks,
    run_doctor,
)
from .init import (
    _default_alias_from_path,
    _ensure_projects_table,
    _prompt_alias,
    run_init,
)
from .onboarding_cmd import chat_id, onboarding_paths
from .topic import run_topic
from .plugins import plugins_cmd
from .run import (
    _default_engine_for_setup,
    _print_version_and_exit,
    _resolve_setup_engine,
    _resolve_transport_id,
    _run_auto_router,
    _setup_needs_config,
    _should_run_interactive,
    _version_callback,
    acquire_config_lock,
    app_main,
    make_engine_cmd,
)
from .config import (
    _CONFIG_PATH_OPTION,
    _config_path_display,
    _exit_config_error,
    _fail_missing_config,
    _flatten_config,
    _load_config_or_exit,
    _normalized_value_from_settings,
    _parse_key_path,
    _parse_value,
    _resolve_config_path_override,
    _toml_literal,
    config_get,
    config_list,
    config_path_cmd,
    config_set,
    config_unset,
)
from .cron import app as cron_app
from .handoff import app as handoff_app
from .reload import reload_command


def _load_settings_optional() -> tuple[TakopiSettings | None, Path | None]:
    try:
        loaded = load_settings_if_exists()
    except ConfigError:
        return None, None
    if loaded is None:
        return None, None
    return loaded


def init(
    alias: str | None = typer.Argument(
        None, help="Project alias (used as /alias in messages)."
    ),
    default: bool = typer.Option(
        False,
        "--default",
        help="Set this project as the default_project.",
    ),
) -> None:
    """Register the current repo as a Takopi project."""
    run_init(
        alias=alias,
        default=default,
        load_or_init_config_fn=load_or_init_config,
        resolve_main_worktree_root_fn=resolve_main_worktree_root,
        resolve_default_base_fn=resolve_default_base,
        list_backend_ids_fn=list_backend_ids,
        resolve_plugins_allowlist_fn=resolve_plugins_allowlist,
    )


def topic_init(
    project: str | None = typer.Argument(
        None, help="Project alias (defaults to current directory name)."
    ),
    branch: str | None = typer.Option(
        None, "--branch", "-b", help="Branch name (defaults to current git branch)."
    ),
) -> None:
    """Create a Telegram topic bound to a project/branch."""
    run_topic(
        project=project,
        branch=branch,
        delete=False,
        config_path=None,
    )


def topic_delete(
    project: str | None = typer.Argument(
        None, help="Project alias (defaults to current directory name)."
    ),
    branch: str | None = typer.Option(
        None, "--branch", "-b", help="Branch name (defaults to current git branch)."
    ),
) -> None:
    """Delete a Telegram topic binding."""
    run_topic(
        project=project,
        branch=branch,
        delete=True,
        config_path=None,
    )


def doctor() -> None:
    """Run configuration checks for the active transport."""
    setup_logging(debug=False, cache_logger_on_first_use=False)
    run_doctor(
        load_settings_fn=load_settings,
        telegram_checks=_doctor_telegram_checks,
        file_checks=_doctor_file_checks,
        voice_checks=_doctor_voice_checks,
    )


def _engine_ids_for_cli() -> list[str]:
    allowlist: list[str] | None = None
    try:
        config, _ = load_or_init_config()
    except ConfigError:
        return list_backend_ids()
    raw_plugins = config.get("plugins")
    if isinstance(raw_plugins, dict):
        enabled = raw_plugins.get("enabled")
        if isinstance(enabled, list):
            allowlist = [
                value.strip()
                for value in enabled
                if isinstance(value, str) and value.strip()
            ]
            if not allowlist:
                allowlist = None
    return list_backend_ids(allowlist=allowlist)


def create_app() -> typer.Typer:
    app = typer.Typer(
        add_completion=False,
        invoke_without_command=True,
        help="Telegram bridge for coding agents. Docs: https://yee88.dev/",
    )
    config_app = typer.Typer(help="Read and modify yee88 config.")
    config_app.command(name="path")(config_path_cmd)
    config_app.command(name="list")(config_list)
    config_app.command(name="get")(config_get)
    config_app.command(name="set")(config_set)
    config_app.command(name="unset")(config_unset)
    topic_app = typer.Typer(help="Manage Telegram topics.")
    topic_app.command(name="init")(topic_init)
    topic_app.command(name="delete")(topic_delete)
    app.command(name="init")(init)
    app.add_typer(topic_app, name="topic")
    app.command(name="chat-id")(chat_id)
    app.command(name="doctor")(doctor)
    app.command(name="onboarding-paths")(onboarding_paths)
    app.command(name="plugins")(plugins_cmd)
    app.add_typer(config_app, name="config")
    app.add_typer(cron_app, name="cron")
    app.add_typer(handoff_app, name="handoff")
    app.command(name="reload")(reload_command)
    app.callback()(app_main)
    for engine_id in _engine_ids_for_cli():
        help_text = f"Run with the {engine_id} engine."
        app.command(name=engine_id, help=help_text)(make_engine_cmd(engine_id))
    return app


def main() -> None:
    app = create_app()
    app()


if __name__ == "__main__":
    main()



import anyio

import pytest

from collections.abc import AsyncIterator

from yee88.model import (
    ActionEvent,
    CompletedEvent,
    ResumeToken,
    StartedEvent,
    TakopiEvent,
)
from yee88.runners.codex import CodexRunner, find_exec_only_flag

CODEX_ENGINE = "codex"


@pytest.mark.anyio
async def test_run_serializes_same_session() -> None:
    runner = CodexRunner(codex_cmd="codex", extra_args=[])
    gate = anyio.Event()
    in_flight = 0
    max_in_flight = 0

    async def run_stub(*_args, **_kwargs) -> AsyncIterator[TakopiEvent]:
        nonlocal in_flight, max_in_flight
        in_flight += 1
        max_in_flight = max(max_in_flight, in_flight)
        try:
            await gate.wait()
            yield CompletedEvent(
                engine=CODEX_ENGINE,
                resume=ResumeToken(engine=CODEX_ENGINE, value="sid"),
                ok=True,
                answer="ok",
            )
        finally:
            in_flight -= 1

    runner.run_impl = run_stub  # type: ignore[assignment]

    async def drain(prompt: str, resume: ResumeToken | None) -> None:
        async for _event in runner.run(prompt, resume):
            pass

    token = ResumeToken(engine=CODEX_ENGINE, value="sid")
    async with anyio.create_task_group() as tg:
        tg.start_soon(drain, "a", token)
        tg.start_soon(drain, "b", token)
        await anyio.sleep(0)
        gate.set()
    assert max_in_flight == 1


@pytest.mark.anyio
async def test_run_allows_parallel_new_sessions() -> None:
    runner = CodexRunner(codex_cmd="codex", extra_args=[])
    gate = anyio.Event()
    in_flight = 0
    max_in_flight = 0

    async def run_stub(*_args, **_kwargs) -> AsyncIterator[TakopiEvent]:
        nonlocal in_flight, max_in_flight
        in_flight += 1
        max_in_flight = max(max_in_flight, in_flight)
        try:
            await gate.wait()
            yield CompletedEvent(
                engine=CODEX_ENGINE,
                resume=ResumeToken(engine=CODEX_ENGINE, value="sid"),
                ok=True,
                answer="ok",
            )
        finally:
            in_flight -= 1

    runner.run_impl = run_stub  # type: ignore[assignment]

    async def drain(prompt: str, resume: ResumeToken | None) -> None:
        async for _event in runner.run(prompt, resume):
            pass

    async with anyio.create_task_group() as tg:
        tg.start_soon(drain, "a", None)
        tg.start_soon(drain, "b", None)
        await anyio.sleep(0)
        gate.set()
    assert max_in_flight == 2


@pytest.mark.anyio
async def test_run_allows_parallel_different_sessions() -> None:
    runner = CodexRunner(codex_cmd="codex", extra_args=[])
    gate = anyio.Event()
    in_flight = 0
    max_in_flight = 0

    async def run_stub(*_args, **_kwargs) -> AsyncIterator[TakopiEvent]:
        nonlocal in_flight, max_in_flight
        in_flight += 1
        max_in_flight = max(max_in_flight, in_flight)
        try:
            await gate.wait()
            yield CompletedEvent(
                engine=CODEX_ENGINE,
                resume=ResumeToken(engine=CODEX_ENGINE, value="sid"),
                ok=True,
                answer="ok",
            )
        finally:
            in_flight -= 1

    runner.run_impl = run_stub  # type: ignore[assignment]

    async def drain(prompt: str, resume: ResumeToken | None) -> None:
        async for _event in runner.run(prompt, resume):
            pass

    token_a = ResumeToken(engine=CODEX_ENGINE, value="sid-a")
    token_b = ResumeToken(engine=CODEX_ENGINE, value="sid-b")
    async with anyio.create_task_group() as tg:
        tg.start_soon(drain, "a", token_a)
        tg.start_soon(drain, "b", token_b)
        await anyio.sleep(0)
        gate.set()
    assert max_in_flight == 2


def test_codex_exec_flags_after_exec() -> None:
    runner = CodexRunner(
        codex_cmd="codex",
        extra_args=["-c", "notify=[]"],
    )
    state = runner.new_state("hi", None)
    args = runner.build_args("hi", None, state=state)
    assert args == [
        "-c",
        "notify=[]",
        "exec",
        "--json",
        "--skip-git-repo-check",
        "--color=never",
        "-",
    ]


@pytest.mark.parametrize(
    ("extra_args", "expected"),
    [
        ([], None),
        (["-c", "notify=[]"], None),
        (["--skip-git-repo-check"], "--skip-git-repo-check"),
        (["--color=never"], "--color=never"),
        (["--output-schema", "schema.json"], "--output-schema"),
        (["--output-last-message=out.txt"], "--output-last-message=out.txt"),
        (["-o", "out.txt"], "-o"),
    ],
)
def test_find_exec_only_flag(extra_args: list[str], expected: str | None) -> None:
    assert find_exec_only_flag(extra_args) == expected


@pytest.mark.anyio
async def test_run_serializes_new_session_after_session_is_known(
    tmp_path, monkeypatch
) -> None:
    gate_path = tmp_path / "gate"
    resume_marker = tmp_path / "resume_started"
    thread_id = "019b73c4-0c3f-7701-a0bb-aac6b4d8a3bc"

    codex_path = tmp_path / "codex"
    codex_path.write_text(
        "#!/usr/bin/env python3\n"
        "import json\n"
        "import os\n"
        "import sys\n"
        "import time\n"
        "\n"
        "gate = os.environ['CODEX_TEST_GATE']\n"
        "resume_marker = os.environ['CODEX_TEST_RESUME_MARKER']\n"
        "thread_id = os.environ['CODEX_TEST_THREAD_ID']\n"
        "\n"
        "args = sys.argv[1:]\n"
        "if 'resume' in args:\n"
        "    print(json.dumps({'type': 'thread.started', 'thread_id': thread_id}), flush=True)\n"
        "    with open(resume_marker, 'w', encoding='utf-8') as f:\n"
        "        f.write('started')\n"
        "        f.flush()\n"
        "    sys.exit(0)\n"
        "\n"
        "print(json.dumps({'type': 'thread.started', 'thread_id': thread_id}), flush=True)\n"
        "while not os.path.exists(gate):\n"
        "    time.sleep(0.001)\n"
        "sys.exit(0)\n",
        encoding="utf-8",
    )
    codex_path.chmod(0o755)

    monkeypatch.setenv("CODEX_TEST_GATE", str(gate_path))
    monkeypatch.setenv("CODEX_TEST_RESUME_MARKER", str(resume_marker))
    monkeypatch.setenv("CODEX_TEST_THREAD_ID", thread_id)

    runner = CodexRunner(codex_cmd=str(codex_path), extra_args=[])

    session_started = anyio.Event()
    resume_value: str | None = None

    new_done = anyio.Event()

    async def run_new() -> None:
        nonlocal resume_value
        async for event in runner.run("hello", None):
            if isinstance(event, StartedEvent):
                resume_value = event.resume.value
                session_started.set()
        new_done.set()

    async def run_resume() -> None:
        assert resume_value is not None
        async for _event in runner.run(
            "resume", ResumeToken(engine=CODEX_ENGINE, value=resume_value)
        ):
            pass

    async with anyio.create_task_group() as tg:
        tg.start_soon(run_new)
        await session_started.wait()

        tg.start_soon(run_resume)
        await anyio.sleep(0.01)

        assert not resume_marker.exists()

        gate_path.write_text("go", encoding="utf-8")
        await new_done.wait()

        with anyio.fail_after(2):
            while not resume_marker.exists():
                await anyio.sleep(0.001)


@pytest.mark.anyio
async def test_codex_runner_preserves_warning_order(tmp_path) -> None:
    thread_id = "019b73c4-0c3f-7701-a0bb-aac6b4d8a3bc"

    codex_path = tmp_path / "codex"
    codex_path.write_text(
        "#!/usr/bin/env python3\n"
        "import json\n"
        "import sys\n"
        "\n"
        "sys.stdin.read()\n"
        "print(json.dumps({'type': 'error', 'message': 'warning one'}), flush=True)\n"
        f"print(json.dumps({{'type': 'thread.started', 'thread_id': '{thread_id}'}}), flush=True)\n"
        "print(json.dumps({'type': 'item.completed', 'item': {'id': 'item_0', 'type': 'agent_message', 'text': 'ok'}}), flush=True)\n",
        encoding="utf-8",
    )
    codex_path.chmod(0o755)

    runner = CodexRunner(codex_cmd=str(codex_path), extra_args=[])
    seen = [evt async for evt in runner.run("hi", None)]

    assert len(seen) == 3
    assert isinstance(seen[0], ActionEvent)
    assert seen[0].phase == "completed"
    assert seen[0].ok is False
    assert seen[0].action.kind == "warning"
    assert seen[0].action.title == "warning one"

    assert isinstance(seen[1], StartedEvent)
    assert seen[1].resume.value == thread_id

    assert isinstance(seen[2], CompletedEvent)
    assert seen[2].resume == seen[1].resume
    assert seen[2].answer == "ok"


@pytest.mark.anyio
async def test_codex_runner_reconnect_notice_is_non_fatal(tmp_path) -> None:
    thread_id = "019b73c4-0c3f-7701-a0bb-aac6b4d8a3bc"

    codex_path = tmp_path / "codex"
    codex_path.write_text(
        "#!/usr/bin/env python3\n"
        "import json\n"
        "import sys\n"
        "\n"
        "sys.stdin.read()\n"
        "print(json.dumps({'type': 'error', 'message': 'Reconnecting... 1/5'}), flush=True)\n"
        f"print(json.dumps({{'type': 'thread.started', 'thread_id': '{thread_id}'}}), flush=True)\n"
        "print(json.dumps({'type': 'item.completed', 'item': {'id': 'item_0', 'type': 'agent_message', 'text': 'ok'}}), flush=True)\n",
        encoding="utf-8",
    )
    codex_path.chmod(0o755)

    runner = CodexRunner(codex_cmd=str(codex_path), extra_args=[])
    seen = [evt async for evt in runner.run("hi", None)]

    assert len(seen) == 3
    assert isinstance(seen[0], ActionEvent)
    assert seen[0].phase == "started"
    assert seen[0].ok is None
    assert seen[0].action.kind == "note"
    assert seen[0].action.title == "Reconnecting... 1/5"

    assert isinstance(seen[1], StartedEvent)
    assert seen[1].resume.value == thread_id

    assert isinstance(seen[2], CompletedEvent)
    assert seen[2].resume == seen[1].resume
    assert seen[2].answer == "ok"


@pytest.mark.anyio
async def test_codex_runner_reconnect_notice_updates_phase(tmp_path) -> None:
    thread_id = "019b73c4-0c3f-7701-a0bb-aac6b4d8a3bc"

    codex_path = tmp_path / "codex"
    codex_path.write_text(
        "#!/usr/bin/env python3\n"
        "import json\n"
        "import sys\n"
        "\n"
        "sys.stdin.read()\n"
        "print(json.dumps({'type': 'error', 'message': 'Reconnecting... 1/5'}), flush=True)\n"
        "print(json.dumps({'type': 'error', 'message': 'Reconnecting... 2/5'}), flush=True)\n"
        f"print(json.dumps({{'type': 'thread.started', 'thread_id': '{thread_id}'}}), flush=True)\n"
        "print(json.dumps({'type': 'item.completed', 'item': {'id': 'item_0', 'type': 'agent_message', 'text': 'ok'}}), flush=True)\n"
        "print(json.dumps({'type': 'turn.completed', 'usage': {'input_tokens': 1, 'cached_input_tokens': 0, 'output_tokens': 1}}), flush=True)\n",
        encoding="utf-8",
    )
    codex_path.chmod(0o755)

    runner = CodexRunner(codex_cmd=str(codex_path), extra_args=[])
    seen = [evt async for evt in runner.run("hi", None)]

    assert len(seen) == 4
    first = seen[0]
    second = seen[1]
    assert isinstance(first, ActionEvent)
    assert isinstance(second, ActionEvent)
    assert first.phase == "started"
    assert second.phase == "updated"
    assert first.action.id == second.action.id == "codex.reconnect"
    assert isinstance(seen[2], StartedEvent)
    assert isinstance(seen[3], CompletedEvent)


@pytest.mark.anyio
async def test_codex_runner_includes_stderr_reason(tmp_path) -> None:
    codex_path = tmp_path / "codex"
    codex_path.write_text(
        "#!/usr/bin/env python3\n"
        "import sys\n"
        "\n"
        "sys.stderr.write('Not inside a trusted directory and --skip-git-repo-check was not specified.\\n')\n"
        "sys.stderr.flush()\n"
        "sys.exit(1)\n",
        encoding="utf-8",
    )
    codex_path.chmod(0o755)

    runner = CodexRunner(codex_cmd=str(codex_path), extra_args=[])
    events = [evt async for evt in runner.run("hi", None)]

    completed = next(evt for evt in events if isinstance(evt, CompletedEvent))
    assert completed.ok is False
    assert completed.error is not None
    assert "codex exec failed (rc=1)." in completed.error
    assert "Not inside a trusted directory" not in completed.error


@pytest.mark.anyio
async def test_run_serializes_two_new_sessions_same_thread(
    tmp_path, monkeypatch
) -> None:
    gate_path = tmp_path / "gate"
    thread_id = "019b73c4-0c3f-7701-a0bb-aac6b4d8a3bc"

    codex_path = tmp_path / "codex"
    codex_path.write_text(
        "#!/usr/bin/env python3\n"
        "import json\n"
        "import os\n"
        "import sys\n"
        "import time\n"
        "\n"
        "gate = os.environ['CODEX_TEST_GATE']\n"
        "thread_id = os.environ['CODEX_TEST_THREAD_ID']\n"
        "\n"
        "print(json.dumps({'type': 'thread.started', 'thread_id': thread_id}), flush=True)\n"
        "while not os.path.exists(gate):\n"
        "    time.sleep(0.001)\n"
        "sys.exit(0)\n",
        encoding="utf-8",
    )
    codex_path.chmod(0o755)

    monkeypatch.setenv("CODEX_TEST_GATE", str(gate_path))
    monkeypatch.setenv("CODEX_TEST_THREAD_ID", thread_id)

    runner = CodexRunner(codex_cmd=str(codex_path), extra_args=[])

    started_first = anyio.Event()
    started_second = anyio.Event()

    async def run_first() -> None:
        async for event in runner.run("one", None):
            if isinstance(event, StartedEvent):
                started_first.set()

    async def run_second() -> None:
        async for event in runner.run("two", None):
            if isinstance(event, StartedEvent):
                started_second.set()

    async with anyio.create_task_group() as tg:
        tg.start_soon(run_first)
        tg.start_soon(run_second)

        with anyio.fail_after(2):
            while not (started_first.is_set() or started_second.is_set()):
                await anyio.sleep(0.001)

        assert not (started_first.is_set() and started_second.is_set())

        gate_path.write_text("go", encoding="utf-8")

        with anyio.fail_after(2):
            await started_first.wait()
            await started_second.wait()



from pathlib import Path

import pytest

from yee88.model import ResumeToken
from yee88.telegram.chat_sessions import ChatSessionStore


@pytest.mark.anyio
async def test_chat_sessions_store_roundtrip(tmp_path) -> None:
    path = tmp_path / "telegram_chat_sessions_state.json"
    store = ChatSessionStore(path)
    await store.set_session_resume(1, None, ResumeToken(engine="codex", value="abc123"))
    await store.set_session_resume(1, 42, ResumeToken(engine="claude", value="res-1"))

    stored_private = await store.get_session_resume(1, None, "codex")
    stored_group = await store.get_session_resume(1, 42, "claude")
    assert stored_private == ResumeToken(engine="codex", value="abc123")
    assert stored_group == ResumeToken(engine="claude", value="res-1")

    store2 = ChatSessionStore(path)
    stored_private_2 = await store2.get_session_resume(1, None, "codex")
    stored_group_2 = await store2.get_session_resume(1, 42, "claude")
    assert stored_private_2 == ResumeToken(engine="codex", value="abc123")
    assert stored_group_2 == ResumeToken(engine="claude", value="res-1")


@pytest.mark.anyio
async def test_chat_sessions_store_clear(tmp_path) -> None:
    path = tmp_path / "telegram_chat_sessions_state.json"
    store = ChatSessionStore(path)
    await store.set_session_resume(2, None, ResumeToken(engine="codex", value="one"))
    await store.set_session_resume(2, 77, ResumeToken(engine="codex", value="two"))

    await store.clear_sessions(2, None)
    assert await store.get_session_resume(2, None, "codex") is None
    assert await store.get_session_resume(2, 77, "codex") == ResumeToken(
        engine="codex",
        value="two",
    )


@pytest.mark.anyio
async def test_chat_sessions_store_drops_sessions_on_cwd_change(
    tmp_path, monkeypatch
) -> None:
    path = tmp_path / "telegram_chat_sessions_state.json"
    dir1 = tmp_path / "dir1"
    dir2 = tmp_path / "dir2"
    dir1.mkdir()
    dir2.mkdir()

    monkeypatch.chdir(dir1)
    store = ChatSessionStore(path)
    await store.set_session_resume(1, None, ResumeToken(engine="codex", value="abc123"))
    assert await store.get_session_resume(1, None, "codex") == ResumeToken(
        engine="codex", value="abc123"
    )

    store2 = ChatSessionStore(path)
    assert await store2.sync_startup_cwd(Path.cwd()) is False
    assert await store2.get_session_resume(1, None, "codex") == ResumeToken(
        engine="codex", value="abc123"
    )

    monkeypatch.chdir(dir2)
    store3 = ChatSessionStore(path)
    assert await store3.sync_startup_cwd(Path.cwd()) is True
    assert await store3.get_session_resume(1, None, "codex") is None



from __future__ import annotations

import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import Any
from collections.abc import Iterable, Mapping

from .backends import EngineBackend
from .config import ConfigError, ProjectsConfig
from .engines import get_backend, list_backend_ids
from .ids import RESERVED_CHAT_COMMANDS
from .logging import get_logger
from .router import AutoRouter, EngineStatus, RunnerEntry
from .settings import TakopiSettings
from .transport_runtime import TransportRuntime

logger = get_logger(__name__)


@dataclass(frozen=True, slots=True)
class RuntimeSpec:
    router: AutoRouter
    projects: ProjectsConfig
    allowlist: list[str] | None
    plugin_configs: Mapping[str, Any] | None
    engine_configs: Mapping[str, Any] | None = None
    watch_config: bool = False

    def to_runtime(self, *, config_path: Path | None) -> TransportRuntime:
        return TransportRuntime(
            router=self.router,
            projects=self.projects,
            allowlist=self.allowlist,
            config_path=config_path,
            plugin_configs=self.plugin_configs,
            engine_configs=self.engine_configs,
            watch_config=self.watch_config,
        )

    def apply(self, runtime: TransportRuntime, *, config_path: Path | None) -> None:
        runtime.update(
            router=self.router,
            projects=self.projects,
            allowlist=self.allowlist,
            config_path=config_path,
            plugin_configs=self.plugin_configs,
            engine_configs=self.engine_configs,
            watch_config=self.watch_config,
        )


def resolve_plugins_allowlist(
    settings: TakopiSettings | None,
) -> list[str] | None:
    if settings is None:
        return None
    enabled = list(settings.plugins.enabled)
    return enabled or None


def resolve_default_engine(
    *,
    override: str | None,
    settings: TakopiSettings,
    config_path: Path,
    engine_ids: list[str],
) -> str:
    default_engine = override or settings.default_engine or "codex"
    if default_engine not in engine_ids:
        available = ", ".join(sorted(engine_ids))
        raise ConfigError(
            f"Unknown default engine {default_engine!r}. Available: {available}."
        )
    return default_engine


def build_router(
    *,
    settings: TakopiSettings,
    config_path: Path,
    backends: list[EngineBackend],
    default_engine: str,
) -> AutoRouter:
    entries: list[RunnerEntry] = []
    warnings: list[str] = []

    for backend in backends:
        engine_id = backend.id
        issue: str | None = None
        status: EngineStatus = "ok"
        engine_cfg: dict
        try:
            engine_cfg = settings.engine_config(engine_id, config_path=config_path)
        except ConfigError as exc:
            if engine_id == default_engine:
                raise
            issue = str(exc)
            status = "bad_config"
            engine_cfg = {}

        try:
            runner = backend.build_runner(engine_cfg, config_path)
        except Exception as exc:
            if engine_id == default_engine:
                raise
            issue = issue or str(exc)
            if engine_cfg:
                try:
                    runner = backend.build_runner({}, config_path)
                except Exception as fallback_exc:  # noqa: BLE001
                    warnings.append(f"{engine_id}: {issue or str(fallback_exc)}")
                    continue
                status = "bad_config"
            else:
                status = "load_error"
                warnings.append(f"{engine_id}: {issue}")
                continue

        cmd = backend.cli_cmd or backend.id
        if shutil.which(cmd) is None:
            status = "missing_cli"
            if issue:
                issue = f"{issue}; {cmd} not found on PATH"
            else:
                issue = f"{cmd} not found on PATH"

        if status != "ok" and engine_id == default_engine:
            raise ConfigError(f"Default engine {engine_id!r} unavailable: {issue}")

        if status != "ok" and engine_id != default_engine:
            warnings.append(f"{engine_id}: {issue}")

        entries.append(
            RunnerEntry(
                engine=engine_id,
                runner=runner,
                status=status,
                issue=issue,
            )
        )

    for warning in warnings:
        logger.warning("setup.warning", issue=warning)

    return AutoRouter(entries=entries, default_engine=default_engine)


def load_backends(
    *,
    engine_ids: list[str],
    allowlist: list[str] | None,
    default_engine: str,
) -> list[EngineBackend]:
    backends: list[EngineBackend] = []
    load_issues: list[str] = []
    for engine_id in engine_ids:
        try:
            backend = get_backend(engine_id, allowlist=allowlist)
        except ConfigError as exc:
            if engine_id == default_engine:
                raise
            load_issues.append(f"{engine_id}: {exc}")
            continue
        backends.append(backend)
    if not backends:
        raise ConfigError("No engine backends are available.")
    for issue in load_issues:
        logger.warning("setup.warning", issue=issue)
    return backends


def build_runtime_spec(
    *,
    settings: TakopiSettings,
    config_path: Path,
    default_engine_override: str | None = None,
    reserved: Iterable[str] = RESERVED_CHAT_COMMANDS,
) -> RuntimeSpec:
    allowlist = resolve_plugins_allowlist(settings)
    engine_ids = list_backend_ids(allowlist=allowlist)
    projects = settings.to_projects_config(
        config_path=config_path,
        engine_ids=engine_ids,
        reserved=reserved,
    )
    default_engine = resolve_default_engine(
        override=default_engine_override,
        settings=settings,
        config_path=config_path,
        engine_ids=engine_ids,
    )
    backends = load_backends(
        engine_ids=engine_ids,
        allowlist=allowlist,
        default_engine=default_engine,
    )
    router = build_router(
        settings=settings,
        config_path=config_path,
        backends=backends,
        default_engine=default_engine,
    )
    return RuntimeSpec(
        router=router,
        projects=projects,
        allowlist=allowlist,
        plugin_configs=settings.plugins.model_extra,
        engine_configs=settings.model_extra,
        watch_config=settings.watch_config,
    )



from __future__ import annotations

from pathlib import Path
import tomllib

from typer.testing import CliRunner

from yee88 import cli
from yee88.config import ConfigError
from yee88.plugins import (
    COMMAND_GROUP,
    ENGINE_GROUP,
    TRANSPORT_GROUP,
    PluginLoadError,
)
from yee88.settings import TakopiSettings
from tests.plugin_fixtures import FakeEntryPoint


def _min_config() -> dict:
    return {
        "transport": "telegram",
        "transports": {"telegram": {"bot_token": "token", "chat_id": 123}},
    }


def test_init_registers_project(monkeypatch, tmp_path: Path) -> None:
    config = _min_config()
    config_path = tmp_path / "yee88.toml"
    repo_path = tmp_path / "repo"
    repo_path.mkdir()
    monkeypatch.chdir(repo_path)

    monkeypatch.setattr(cli, "load_or_init_config", lambda: (config, config_path))
    monkeypatch.setattr(cli, "resolve_main_worktree_root", lambda _path: None)
    monkeypatch.setattr(cli, "resolve_default_base", lambda _path: "main")
    monkeypatch.setattr(cli, "list_backend_ids", lambda allowlist=None: ["codex"])
    monkeypatch.setattr(cli, "resolve_plugins_allowlist", lambda _settings: None)
    monkeypatch.setattr(cli.typer, "prompt", lambda *args, **kwargs: "demo")

    runner = CliRunner()
    result = runner.invoke(cli.create_app(), ["init", "--default"])

    assert result.exit_code == 0
    assert "saved project" in result.output

    data = tomllib.loads(config_path.read_text(encoding="utf-8"))
    project = data["projects"]["demo"]
    assert project["path"] == str(repo_path)
    assert project["default_engine"] == "codex"
    assert project["worktrees_dir"] == ".worktrees"
    assert project["worktree_base"] == "main"
    assert data["default_project"] == "demo"


def test_init_declines_overwrite(monkeypatch, tmp_path: Path) -> None:
    config = _min_config()
    config["projects"] = {"demo": {"path": "/tmp/repo"}}
    config_path = tmp_path / "yee88.toml"
    repo_path = tmp_path / "repo"
    repo_path.mkdir()
    monkeypatch.chdir(repo_path)

    monkeypatch.setattr(cli, "load_or_init_config", lambda: (config, config_path))
    monkeypatch.setattr(cli, "resolve_main_worktree_root", lambda _path: None)
    monkeypatch.setattr(cli, "resolve_default_base", lambda _path: None)
    monkeypatch.setattr(cli, "list_backend_ids", lambda allowlist=None: ["codex"])
    monkeypatch.setattr(cli, "resolve_plugins_allowlist", lambda _settings: None)
    monkeypatch.setattr(cli.typer, "confirm", lambda *args, **kwargs: False)

    runner = CliRunner()
    result = runner.invoke(cli.create_app(), ["init", "demo"])

    assert result.exit_code == 1


def test_plugins_cmd_loads_and_reports_errors(monkeypatch) -> None:
    entrypoints = {
        ENGINE_GROUP: [
            FakeEntryPoint(
                "codex",
                "yee88.runners.codex:BACKEND",
                ENGINE_GROUP,
                dist_name="yee88",
            ),
            FakeEntryPoint(
                "broken",
                "yee88.runners.broken:BACKEND",
                ENGINE_GROUP,
                dist_name="yee88",
            ),
        ],
        TRANSPORT_GROUP: [
            FakeEntryPoint(
                "telegram",
                "yee88.transports.telegram:BACKEND",
                TRANSPORT_GROUP,
                dist_name="yee88",
            )
        ],
        COMMAND_GROUP: [
            FakeEntryPoint(
                "hello",
                "yee88.commands.hello:BACKEND",
                COMMAND_GROUP,
                dist_name="thirdparty",
            )
        ],
    }

    def _list_entrypoints(group: str, reserved_ids=None):
        _ = reserved_ids
        return entrypoints[group]

    calls: list[tuple[str, str]] = []

    def _get_backend(name: str, allowlist=None):
        _ = allowlist
        calls.append(("engine", name))
        if name == "broken":
            raise ConfigError("boom")
        return object()

    def _get_transport(name: str, allowlist=None):
        _ = allowlist
        calls.append(("transport", name))
        return object()

    def _get_command(name: str, allowlist=None):
        _ = allowlist
        calls.append(("command", name))
        return object()

    monkeypatch.setattr(cli, "_load_settings_optional", lambda: (None, None))
    monkeypatch.setattr(cli, "resolve_plugins_allowlist", lambda _settings: ["yee88"])
    monkeypatch.setattr(cli, "list_entrypoints", _list_entrypoints)
    monkeypatch.setattr(cli, "get_backend", _get_backend)
    monkeypatch.setattr(cli, "get_transport", _get_transport)
    monkeypatch.setattr(cli, "get_command", _get_command)
    monkeypatch.setattr(
        cli,
        "get_load_errors",
        lambda: [
            PluginLoadError(
                ENGINE_GROUP,
                "broken",
                "yee88.runners.broken:BACKEND",
                "yee88",
                "boom",
            ),
            PluginLoadError(
                TRANSPORT_GROUP,
                "wire",
                "yee88.transports.wire:BACKEND",
                "yee88",
                "missing",
            ),
            PluginLoadError(
                COMMAND_GROUP,
                "hello",
                "yee88.commands.hello:BACKEND",
                "thirdparty",
                "oops",
            ),
        ],
    )

    runner = CliRunner()
    result = runner.invoke(cli.create_app(), ["plugins", "--load"])

    assert result.exit_code == 0
    assert "engine backends:" in result.output
    assert "transport backends:" in result.output
    assert "command backends:" in result.output
    assert "codex (yee88) enabled" in result.output
    assert "hello (thirdparty) disabled" in result.output
    assert "errors:" in result.output
    assert "engine broken (yee88): boom" in result.output
    assert "transport wire (yee88): missing" in result.output
    assert "command hello (thirdparty): oops" in result.output

    assert ("engine", "codex") in calls
    assert ("engine", "broken") in calls
    assert ("transport", "telegram") in calls
    assert ("command", "hello") not in calls


def test_onboarding_paths_calls_debug(monkeypatch) -> None:
    called = {"count": 0}

    def _debug() -> None:
        called["count"] += 1

    monkeypatch.setattr(cli.onboarding, "debug_onboarding_paths", _debug)

    runner = CliRunner()
    result = runner.invoke(cli.create_app(), ["onboarding-paths"])

    assert result.exit_code == 0
    assert called["count"] == 1


def test_config_path_cmd_outputs_override(tmp_path: Path) -> None:
    config_path = tmp_path / "yee88.toml"
    runner = CliRunner()
    result = runner.invoke(
        cli.create_app(),
        ["config", "path", "--config-path", str(config_path)],
    )

    assert result.exit_code == 0
    assert result.output.strip() == str(config_path)


def test_config_path_cmd_defaults_to_home(monkeypatch, tmp_path: Path) -> None:
    monkeypatch.setenv("HOME", str(tmp_path))
    config_path = tmp_path / ".yee88" / "yee88.toml"
    monkeypatch.setattr(cli, "HOME_CONFIG_PATH", config_path)

    runner = CliRunner()
    result = runner.invoke(cli.create_app(), ["config", "path"])

    assert result.exit_code == 0
    assert result.output.strip() == "~/.yee88/yee88.toml"


def test_doctor_rejects_non_telegram_transport(monkeypatch) -> None:
    settings = TakopiSettings.model_validate(
        {
            "transport": "local",
            "transports": {"telegram": {"bot_token": "token", "chat_id": 123}},
        }
    )
    monkeypatch.setattr(cli, "load_settings", lambda: (settings, Path("x")))

    runner = CliRunner()
    result = runner.invoke(cli.create_app(), ["doctor"])

    assert result.exit_code == 1
    assert "telegram transport only" in result.output



from __future__ import annotations

from collections.abc import Iterable

from .backends import EngineBackend
from .config import ConfigError
from .plugins import ENGINE_GROUP, list_ids, load_plugin_backend
from .ids import RESERVED_ENGINE_IDS


def _validate_engine_backend(backend: object, ep) -> None:
    if not isinstance(backend, EngineBackend):
        raise TypeError(f"{ep.value} is not an EngineBackend")
    if backend.id != ep.name:
        raise ValueError(
            f"{ep.value} engine id {backend.id!r} does not match entrypoint {ep.name!r}"
        )


def get_backend(
    engine_id: str, *, allowlist: Iterable[str] | None = None
) -> EngineBackend:
    if engine_id.lower() in RESERVED_ENGINE_IDS:
        raise ConfigError(f"Engine id {engine_id!r} is reserved.")
    backend = load_plugin_backend(
        ENGINE_GROUP,
        engine_id,
        allowlist=allowlist,
        validator=_validate_engine_backend,
        kind_label="engine",
    )
    assert backend is not None
    return backend


def list_backends(*, allowlist: Iterable[str] | None = None) -> list[EngineBackend]:
    backends: list[EngineBackend] = []
    for engine_id in list_backend_ids(allowlist=allowlist):
        try:
            backends.append(get_backend(engine_id, allowlist=allowlist))
        except ConfigError:
            continue
    if not backends:
        raise ConfigError("No engine backends are available.")
    return backends


def list_backend_ids(*, allowlist: Iterable[str] | None = None) -> list[str]:
    return list_ids(
        ENGINE_GROUP,
        allowlist=allowlist,
        reserved_ids=RESERVED_ENGINE_IDS,
    )



from __future__ import annotations

import re
from datetime import datetime, timedelta
from pathlib import Path

import typer

from ..config import HOME_CONFIG_PATH
from ..cron.manager import CronManager
from ..cron.models import CronJob
from ..settings import load_settings_if_exists
from ..engines import list_backend_ids

app = typer.Typer(help="Manage yee88 cron jobs")


def get_cron_manager() -> CronManager:
    return CronManager(HOME_CONFIG_PATH.parent)


def _validate_project(project: str) -> None:
    if not project:
        return
    result = load_settings_if_exists()
    if result is None:
        raise ValueError(f"未找到配置文件，无法验证项目: {project}")
    settings, config_path = result
    engine_ids = list_backend_ids()
    projects_config = settings.to_projects_config(config_path=config_path, engine_ids=engine_ids)
    if project.lower() not in projects_config.projects:
        available = list(projects_config.projects.keys())
        if available:
            raise ValueError(f"未知项目: {project}。可用项目: {', '.join(available)}")
        else:
            raise ValueError(f"未知项目: {project}。请先使用 'yee88 init' 注册项目")


def _parse_one_time(schedule: str) -> str:
    """解析一次性任务时间，支持相对时间和 ISO 8601 格式。"""
    now = datetime.now()

    # 相对时间格式: +30s, +5m, +2h, +1d
    if schedule.startswith("+"):
        match = re.match(r"\+(\d+)([smhd])", schedule)
        if not match:
            raise ValueError(
                f"无效的时间格式: {schedule}。使用 +30s, +5m, +2h, +1d 或 ISO 8601 (2026-02-01T10:00:00)"
            )

        value, unit = int(match.group(1)), match.group(2)
        delta = {
            "s": timedelta(seconds=value),
            "m": timedelta(minutes=value),
            "h": timedelta(hours=value),
            "d": timedelta(days=value),
        }[unit]

        return (now + delta).isoformat()

    # ISO 8601 格式
    try:
        dt = datetime.fromisoformat(schedule)
        if dt <= now:
            raise ValueError("执行时间必须在未来")
        return dt.isoformat()
    except ValueError as e:
        if "执行时间必须在未来" in str(e):
            raise
        raise ValueError(
            f"无效的时间格式: {schedule}。使用 +30s, +5m, +2h, +1d 或 ISO 8601 (2026-02-01T10:00:00)"
        )


@app.command()
def add(
    id: str = typer.Argument(...),
    schedule: str = typer.Argument(...),
    message: str = typer.Argument(...),
    project: str = typer.Option("", "--project", "-p", help="项目别名（可选，如 takopi）"),
    one_time: bool = typer.Option(False, "--one-time", "-o", help="一次性任务，执行后自动删除"),
):
    try:
        manager = get_cron_manager()
        manager.load()

        _validate_project(project)

        if one_time:
            schedule = _parse_one_time(schedule)

        job = CronJob(
            id=id,
            schedule=schedule,
            message=message,
            project=project,
            enabled=True,
            one_time=one_time,
        )

        manager.add(job)

        if one_time:
            typer.echo(f"✅ 已添加一次性任务: {id}")
            typer.echo(f"   执行时间: {schedule[:19]}")
        else:
            typer.echo(f"✅ 已添加定时任务: {id}")
            typer.echo(f"   时间: {schedule}")
        if project:
            typer.echo(f"   项目: {project}")
        typer.echo(f"   消息: {message}")

    except ValueError as e:
        typer.echo(f"❌ 错误: {e}", err=True)
        raise typer.Exit(1)


@app.command()
def list(
    show_all: bool = typer.Option(False, "--all", "-a"),
):
    manager = get_cron_manager()
    manager.load()

    jobs = manager.list()

    if not jobs:
        typer.echo("暂无定时任务")
        return

    if not show_all:
        jobs = [j for j in jobs if j.enabled]

    typer.echo(f"{'ID':<20} {'TYPE':<8} {'SCHEDULE':<20} {'STATUS':<10} {'PROJECT'}")
    typer.echo("-" * 90)

    for job in jobs:
        status = "✓ enabled" if job.enabled else "✗ disabled"
        job_type = "once" if job.one_time else "cron"
        schedule_display = job.schedule[:19] if job.one_time else job.schedule
        if len(schedule_display) > 20:
            schedule_display = schedule_display[:17] + "..."
        project_display = job.project
        if len(project_display) > 25:
            project_display = "..." + project_display[-22:]
        typer.echo(f"{job.id:<20} {job_type:<8} {schedule_display:<20} {status:<10} {project_display}")


@app.command()
def enable(
    id: str = typer.Argument(...),
):
    manager = get_cron_manager()
    manager.load()

    if manager.enable(id):
        typer.echo(f"✅ 已启用: {id}")
    else:
        typer.echo(f"❌ 未找到任务: {id}", err=True)
        raise typer.Exit(1)


@app.command()
def disable(
    id: str = typer.Argument(...),
):
    manager = get_cron_manager()
    manager.load()

    if manager.disable(id):
        typer.echo(f"⏸️  已禁用: {id}")
    else:
        typer.echo(f"❌ 未找到任务: {id}", err=True)
        raise typer.Exit(1)


@app.command()
def remove(
    id: str = typer.Argument(...),
    force: bool = typer.Option(False, "--force", "-f"),
):
    manager = get_cron_manager()
    manager.load()

    if not force:
        confirm = typer.confirm(f"确定要删除任务 '{id}' 吗？")
        if not confirm:
            typer.echo("已取消")
            raise typer.Exit(0)

    if manager.remove(id):
        typer.echo(f"🗑️  已删除: {id}")
    else:
        typer.echo(f"❌ 未找到任务: {id}", err=True)
        raise typer.Exit(1)


@app.command()
def run(
    id: str = typer.Argument(...),
):
    manager = get_cron_manager()
    manager.load()

    job = manager.get(id)
    if not job:
        typer.echo(f"❌ 未找到任务: {id}", err=True)
        raise typer.Exit(1)

    typer.echo(f"🚀 执行任务: {id}")
    typer.echo(f"   路径: {job.project}")
    typer.echo(f"   消息: {job.message}")
    typer.echo(f"   计划时间: {job.schedule}")
    typer.echo("✅ 测试执行完成")



from __future__ import annotations

from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Literal, Protocol, overload, runtime_checkable

from .config import ConfigError
from .context import RunContext
from .ids import RESERVED_COMMAND_IDS
from .model import EngineId
from .plugins import COMMAND_GROUP, list_ids, load_plugin_backend
from .transport import MessageRef, RenderedMessage
from .transport_runtime import TransportRuntime

RunMode = Literal["emit", "capture"]


@dataclass(frozen=True, slots=True)
class RunRequest:
    prompt: str
    engine: EngineId | None = None
    context: RunContext | None = None


@dataclass(frozen=True, slots=True)
class RunResult:
    engine: EngineId
    message: RenderedMessage | None


class CommandExecutor(Protocol):
    async def send(
        self,
        message: RenderedMessage | str,
        *,
        reply_to: MessageRef | None = None,
        notify: bool = True,
    ) -> MessageRef | None: ...

    async def run_one(
        self, request: RunRequest, *, mode: RunMode = "emit"
    ) -> RunResult: ...

    async def run_many(
        self,
        requests: Sequence[RunRequest],
        *,
        mode: RunMode = "emit",
        parallel: bool = False,
    ) -> list[RunResult]: ...


@dataclass(frozen=True, slots=True)
class CommandContext:
    command: str
    text: str
    args_text: str
    args: tuple[str, ...]
    message: MessageRef
    reply_to: MessageRef | None
    reply_text: str | None
    config_path: Path | None
    plugin_config: dict[str, Any]
    runtime: TransportRuntime
    executor: CommandExecutor


@dataclass(frozen=True, slots=True)
class CommandResult:
    text: str
    notify: bool = True
    reply_to: MessageRef | None = None


@runtime_checkable
class CommandBackend(Protocol):
    id: str
    description: str

    async def handle(self, ctx: CommandContext) -> CommandResult | None: ...


def _validate_command_backend(backend: object, ep) -> None:
    if not isinstance(backend, CommandBackend):
        raise TypeError(f"{ep.value} is not a CommandBackend")
    if backend.id != ep.name:
        raise ValueError(
            f"{ep.value} command id {backend.id!r} does not match entrypoint {ep.name!r}"
        )


@overload
def get_command(
    command_id: str,
    *,
    allowlist: Iterable[str] | None = None,
    required: Literal[True] = True,
) -> CommandBackend: ...


@overload
def get_command(
    command_id: str,
    *,
    allowlist: Iterable[str] | None = None,
    required: Literal[False],
) -> CommandBackend | None: ...


def get_command(
    command_id: str,
    *,
    allowlist: Iterable[str] | None = None,
    required: bool = True,
) -> CommandBackend | None:
    if command_id.lower() in RESERVED_COMMAND_IDS:
        raise ConfigError(f"Command id {command_id!r} is reserved.")
    return load_plugin_backend(
        COMMAND_GROUP,
        command_id,
        allowlist=allowlist,
        validator=_validate_command_backend,
        kind_label="command",
        required=required,
    )


def list_command_ids(*, allowlist: Iterable[str] | None = None) -> list[str]:
    return list_ids(
        COMMAND_GROUP,
        allowlist=allowlist,
        reserved_ids=RESERVED_COMMAND_IDS,
    )



import uuid

import anyio
import pytest

from yee88.runner_bridge import ExecBridgeConfig, IncomingMessage, handle_message
from yee88.markdown import MarkdownParts, MarkdownPresenter
from yee88.model import ResumeToken, TakopiEvent
from yee88.telegram.render import prepare_telegram
from yee88.runners.codex import CodexRunner
from yee88.runners.mock import Advance, Emit, Raise, Return, ScriptRunner, Wait
from yee88.settings import load_settings, require_telegram
from yee88.transport import MessageRef, RenderedMessage, SendOptions
from tests.factories import action_completed, action_started

CODEX_ENGINE = "codex"


class FakeTransport:
    def __init__(self) -> None:
        self._next_id = 1
        self.send_calls: list[dict] = []
        self.edit_calls: list[dict] = []
        self.delete_calls: list[MessageRef] = []

    async def send(
        self,
        *,
        channel_id: int | str,
        message: RenderedMessage,
        options: SendOptions | None = None,
    ) -> MessageRef:
        ref = MessageRef(channel_id=channel_id, message_id=self._next_id)
        self._next_id += 1
        self.send_calls.append(
            {
                "ref": ref,
                "channel_id": channel_id,
                "message": message,
                "options": options,
            }
        )
        return ref

    async def edit(
        self, *, ref: MessageRef, message: RenderedMessage, wait: bool = True
    ) -> MessageRef:
        self.edit_calls.append({"ref": ref, "message": message, "wait": wait})
        return ref

    async def delete(self, *, ref: MessageRef) -> bool:
        self.delete_calls.append(ref)
        return True

    async def close(self) -> None:
        return None


class _FakeClock:
    def __init__(self, start: float = 0.0) -> None:
        self._now = start

    def __call__(self) -> float:
        return self._now

    def set(self, value: float) -> None:
        self._now = value


def _return_runner(
    *, answer: str = "ok", resume_value: str | None = None
) -> ScriptRunner:
    return ScriptRunner(
        [Return(answer=answer)],
        engine=CODEX_ENGINE,
        resume_value=resume_value,
    )


def test_require_telegram_rejects_empty_token(tmp_path) -> None:
    from yee88.config import ConfigError

    config_path = tmp_path / "yee88.toml"
    config_path.write_text(
        'transport = "telegram"\n\n[transports.telegram]\n'
        'bot_token = "   "\nchat_id = 123\n',
        encoding="utf-8",
    )

    with pytest.raises(ConfigError, match="bot_token"):
        settings, _ = load_settings(config_path)
        require_telegram(settings, config_path)


def test_load_settings_rejects_string_chat_id(tmp_path) -> None:
    from yee88.config import ConfigError

    config_path = tmp_path / "yee88.toml"
    config_path.write_text(
        'transport = "telegram"\n\n[transports.telegram]\n'
        'bot_token = "token"\nchat_id = "123"\n',
        encoding="utf-8",
    )

    with pytest.raises(ConfigError, match="chat_id"):
        load_settings(config_path)


def test_codex_extract_resume_finds_command() -> None:
    uuid = "019b66fc-64c2-7a71-81cd-081c504cfeb2"
    runner = CodexRunner(codex_cmd="codex", extra_args=[])
    text = f"`codex resume {uuid}`"

    assert runner.extract_resume(text) == ResumeToken(engine=CODEX_ENGINE, value=uuid)


def test_codex_extract_resume_uses_last_resume_line() -> None:
    uuid_first = "019b66fc-64c2-7a71-81cd-081c504cfeb2"
    uuid_last = "123e4567-e89b-12d3-a456-426614174000"
    runner = CodexRunner(codex_cmd="codex", extra_args=[])
    text = f"`codex resume {uuid_first}`\n\n`codex resume {uuid_last}`"

    assert runner.extract_resume(text) == ResumeToken(
        engine=CODEX_ENGINE, value=uuid_last
    )


def test_codex_extract_resume_ignores_malformed_resume_line() -> None:
    runner = CodexRunner(codex_cmd="codex", extra_args=[])
    text = "codex resume"

    assert runner.extract_resume(text) is None


def test_codex_extract_resume_accepts_plain_line() -> None:
    uuid = "019b66fc-64c2-7a71-81cd-081c504cfeb2"
    runner = CodexRunner(codex_cmd="codex", extra_args=[])
    text = f"codex resume {uuid}"

    assert runner.extract_resume(text) == ResumeToken(engine=CODEX_ENGINE, value=uuid)


def test_codex_extract_resume_accepts_uuid7() -> None:
    uuid7 = getattr(uuid, "uuid7", None)
    assert uuid7 is not None
    token = str(uuid7())
    runner = CodexRunner(codex_cmd="codex", extra_args=[])
    text = f"`codex resume {token}`"

    assert runner.extract_resume(text) == ResumeToken(engine=CODEX_ENGINE, value=token)


def test_prepare_telegram_trims_body_preserves_footer() -> None:
    body_limit = 3500
    parts = MarkdownParts(
        header="header",
        body="x" * (body_limit + 100),
        footer="footer",
    )

    rendered, _ = prepare_telegram(parts)

    chunks = [chunk for chunk in rendered.split("\n\n") if chunk]
    assert chunks[0] == "header"
    assert chunks[-1].rstrip() == "footer"
    assert len(chunks[1]) == body_limit
    assert chunks[1].endswith("…")


def test_prepare_telegram_preserves_entities_on_truncate() -> None:
    body_limit = 3500
    parts = MarkdownParts(
        header="h",
        body="**bold** " + ("x" * (body_limit + 100)),
    )

    _, entities = prepare_telegram(parts)

    assert any(e.get("type") == "bold" for e in entities)


@pytest.mark.anyio
async def test_final_notify_sends_loud_final_message() -> None:
    transport = FakeTransport()
    runner = _return_runner(answer="ok")
    cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )

    await handle_message(
        cfg,
        runner=runner,
        incoming=IncomingMessage(channel_id=123, message_id=10, text="hi"),
        resume_token=None,
    )

    assert len(transport.send_calls) == 2
    assert transport.send_calls[0]["options"].notify is False
    assert transport.send_calls[1]["options"].notify is True


@pytest.mark.anyio
async def test_handle_message_strips_resume_line_from_prompt() -> None:
    transport = FakeTransport()
    runner = ScriptRunner([Return(answer="ok")], engine=CODEX_ENGINE)
    cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    resume = ResumeToken(engine=CODEX_ENGINE, value="sid")
    text = "do this\n`codex resume sid`\nand that"

    await handle_message(
        cfg,
        runner=runner,
        incoming=IncomingMessage(channel_id=123, message_id=10, text=text),
        resume_token=resume,
    )

    assert runner.calls
    prompt, passed_resume = runner.calls[0]
    assert prompt == "do this\nand that"
    assert passed_resume == resume


@pytest.mark.anyio
async def test_long_final_message_edits_progress_message() -> None:
    transport = FakeTransport()
    runner = _return_runner(answer="x" * 10_000)
    cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=False,
    )

    await handle_message(
        cfg,
        runner=runner,
        incoming=IncomingMessage(channel_id=123, message_id=10, text="hi"),
        resume_token=None,
    )

    assert len(transport.send_calls) == 1
    assert transport.send_calls[0]["options"].notify is False
    assert transport.edit_calls
    assert "done" in transport.edit_calls[-1]["message"].text.lower()


@pytest.mark.anyio
async def test_progress_edits_are_best_effort() -> None:
    transport = FakeTransport()
    clock = _FakeClock()
    events: list[TakopiEvent] = [
        action_started("item_0", "command", "echo 1"),
        action_started("item_1", "command", "echo 2"),
    ]
    runner = ScriptRunner(
        [
            Emit(events[0], at=0.2),
            Emit(events[1], at=0.4),
            Advance(1.0),
            Return(answer="ok"),
        ],
        engine=CODEX_ENGINE,
        advance=clock.set,
    )
    cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )

    await handle_message(
        cfg,
        runner=runner,
        incoming=IncomingMessage(channel_id=123, message_id=10, text="hi"),
        resume_token=None,
        clock=clock,
    )

    assert transport.edit_calls
    assert all(call["wait"] is False for call in transport.edit_calls)
    assert "working" in transport.edit_calls[-1]["message"].text.lower()


@pytest.mark.anyio
async def test_bridge_flow_sends_progress_edits_and_final_resume() -> None:
    transport = FakeTransport()
    clock = _FakeClock()
    events: list[TakopiEvent] = [
        action_started("item_0", "command", "echo ok"),
        action_completed(
            "item_0",
            "command",
            "echo ok",
            ok=True,
            detail={"exit_code": 0},
        ),
    ]
    session_id = "123e4567-e89b-12d3-a456-426614174000"
    runner = ScriptRunner(
        [
            Emit(events[0], at=0.0),
            Emit(events[1], at=2.1),
            Return(answer="done"),
        ],
        engine=CODEX_ENGINE,
        advance=clock.set,
        resume_value=session_id,
    )
    cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )

    await handle_message(
        cfg,
        runner=runner,
        incoming=IncomingMessage(channel_id=123, message_id=42, text="do it"),
        resume_token=None,
        clock=clock,
    )

    assert transport.send_calls[0]["options"].reply_to.message_id == 42
    assert "starting" in transport.send_calls[0]["message"].text
    assert "codex" in transport.send_calls[0]["message"].text
    assert len(transport.edit_calls) >= 1
    assert session_id in transport.send_calls[-1]["message"].text
    assert "codex resume" in transport.send_calls[-1]["message"].text.lower()
    assert transport.send_calls[-1]["options"].replace == transport.send_calls[0]["ref"]


@pytest.mark.anyio
async def test_final_message_includes_ctx_line() -> None:
    transport = FakeTransport()
    clock = _FakeClock()
    session_id = "123e4567-e89b-12d3-a456-426614174000"
    runner = ScriptRunner(
        [Return(answer="done")],
        engine=CODEX_ENGINE,
        resume_value=session_id,
    )
    cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )

    await handle_message(
        cfg,
        runner=runner,
        incoming=IncomingMessage(channel_id=123, message_id=42, text="do it"),
        resume_token=None,
        context_line="`ctx: yee88 @feat/api`",
        clock=clock,
    )

    assert transport.send_calls
    final_text = transport.send_calls[-1]["message"].text
    assert "`ctx: yee88 @feat/api`" in final_text
    assert "codex resume" in final_text.lower()


@pytest.mark.anyio
async def test_handle_message_cancelled_renders_cancelled_state() -> None:
    transport = FakeTransport()
    session_id = "019b66fc-64c2-7a71-81cd-081c504cfeb2"
    hold = anyio.Event()
    runner = ScriptRunner(
        [Wait(hold)],
        engine=CODEX_ENGINE,
        resume_value=session_id,
    )
    cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )
    running_tasks: dict = {}

    async def run_handle_message() -> None:
        await handle_message(
            cfg,
            runner=runner,
            incoming=IncomingMessage(
                channel_id=123, message_id=10, text="do something"
            ),
            resume_token=None,
            running_tasks=running_tasks,
        )

    async with anyio.create_task_group() as tg:
        tg.start_soon(run_handle_message)
        for _ in range(100):
            if running_tasks:
                break
            await anyio.sleep(0)
        assert running_tasks
        running_task = running_tasks[next(iter(running_tasks))]
        with anyio.fail_after(1):
            await running_task.resume_ready.wait()
        running_task.cancel_requested.set()

    assert len(transport.send_calls) == 1  # Progress message
    assert len(transport.edit_calls) >= 1
    last_edit = transport.edit_calls[-1]["message"].text
    assert "cancelled" in last_edit.lower()
    assert session_id in last_edit


@pytest.mark.anyio
async def test_handle_message_error_preserves_resume_token() -> None:
    transport = FakeTransport()
    session_id = "019b66fc-64c2-7a71-81cd-081c504cfeb2"
    runner = ScriptRunner(
        [Raise(RuntimeError("boom"))],
        engine=CODEX_ENGINE,
        resume_value=session_id,
    )
    cfg = ExecBridgeConfig(
        transport=transport,
        presenter=MarkdownPresenter(),
        final_notify=True,
    )

    await handle_message(
        cfg,
        runner=runner,
        incoming=IncomingMessage(channel_id=123, message_id=10, text="do something"),
        resume_token=None,
    )

    assert transport.edit_calls
    last_edit = transport.edit_calls[-1]["message"].text
    assert "error" in last_edit.lower()
    assert session_id in last_edit
    assert "codex resume" in last_edit.lower()



from __future__ import annotations

from collections.abc import Awaitable, Callable, Sequence
from typing import TYPE_CHECKING

from ...context import RunContext
from ...directives import DirectiveError
from ...transport_runtime import ResolvedMessage
from ..context import _merge_topic_context
from ..files import parse_file_command
from ..topic_state import TopicStateStore
from ..topics import _topic_key, _topics_chat_project
from ..types import TelegramIncomingMessage
from .file_transfer import (
    FILE_PUT_USAGE,
    _format_file_put_failures,
    _handle_file_put_group,
    _save_file_put_group,
)
from .parse import _parse_slash_command
from .reply import make_reply

if TYPE_CHECKING:
    from ..bridge import TelegramBridgeConfig


async def _handle_media_group(
    cfg: TelegramBridgeConfig,
    messages: Sequence[TelegramIncomingMessage],
    topic_store: TopicStateStore | None,
    run_prompt: Callable[
        [TelegramIncomingMessage, str, ResolvedMessage], Awaitable[None]
    ]
    | None = None,
    resolve_prompt: Callable[
        [TelegramIncomingMessage, str, RunContext | None],
        Awaitable[ResolvedMessage | None],
    ]
    | None = None,
) -> None:
    if not messages:
        return
    ordered = sorted(messages, key=lambda item: item.message_id)
    command_msg = next(
        (item for item in ordered if item.text.strip()),
        ordered[0],
    )
    reply = make_reply(cfg, command_msg)
    topic_key = _topic_key(command_msg, cfg) if topic_store is not None else None
    chat_project = _topics_chat_project(cfg, command_msg.chat_id)
    bound_context = (
        await topic_store.get_context(*topic_key)
        if topic_store is not None and topic_key is not None
        else None
    )
    ambient_context = _merge_topic_context(
        chat_project=chat_project, bound=bound_context
    )
    command_id, args_text = _parse_slash_command(command_msg.text)
    if command_id == "file":
        command, rest, error = parse_file_command(args_text)
        if error is not None:
            await reply(text=error)
            return
        if command == "put":
            await _handle_file_put_group(
                cfg,
                command_msg,
                rest,
                ordered,
                ambient_context,
                topic_store,
            )
            return
    if cfg.files.enabled and cfg.files.auto_put:
        caption_text = command_msg.text.strip()
        if cfg.files.auto_put_mode == "prompt" and caption_text:
            if resolve_prompt is None:
                try:
                    resolved = cfg.runtime.resolve_message(
                        text=caption_text,
                        reply_text=command_msg.reply_to_text,
                        ambient_context=ambient_context,
                        chat_id=command_msg.chat_id,
                    )
                except DirectiveError as exc:
                    await reply(text=f"error:\n{exc}")
                    return
            else:
                resolved = await resolve_prompt(
                    command_msg, caption_text, ambient_context
                )
            if resolved is None:
                return
            saved_group = await _save_file_put_group(
                cfg,
                command_msg,
                "",
                ordered,
                resolved.context,
                topic_store,
            )
            if saved_group is None:
                return
            if not saved_group.saved:
                failure_text = _format_file_put_failures(saved_group.failed)
                text = "failed to upload files."
                if failure_text is not None:
                    text = f"{text}\n\n{failure_text}"
                await reply(text=text)
                return
            if saved_group.failed:
                failure_text = _format_file_put_failures(saved_group.failed)
                if failure_text is not None:
                    await reply(text=f"some files failed to upload.\n\n{failure_text}")
            if run_prompt is None:
                await reply(text=FILE_PUT_USAGE)
                return
            paths = [
                item.rel_path.as_posix()
                for item in saved_group.saved
                if item.rel_path is not None
            ]
            files_text = "\n".join(f"- {path}" for path in paths)
            prompt_base = resolved.prompt
            annotation = f"[uploaded files]\n{files_text}"
            if prompt_base and prompt_base.strip():
                prompt = f"{prompt_base}\n\n{annotation}"
            else:
                prompt = annotation
            await run_prompt(command_msg, prompt, resolved)
            return
        if not caption_text:
            await _handle_file_put_group(
                cfg,
                command_msg,
                "",
                ordered,
                ambient_context,
                topic_store,
            )
            return
    await reply(text=FILE_PUT_USAGE)



from __future__ import annotations

from pathlib import Path
from typing import Any, cast

import pytest
from rich.console import Console
from rich.table import Table
from rich.text import Text

from yee88.config import ConfigError
from yee88.telegram import onboarding
from yee88.telegram.api_models import Chat, Message, Update, User
from yee88.telegram.client import TelegramRetryAfter


class DummyUI:
    def __init__(
        self,
        *,
        confirms: list[bool | None] | None = None,
        selects: list[Any] | None = None,
        passwords: list[str | None] | None = None,
    ) -> None:
        self.confirms = iter(confirms or [])
        self.selects = iter(selects or [])
        self.passwords = iter(passwords or [])
        self.printed: list[object] = []
        self.steps: list[tuple[str, int]] = []

    def panel(
        self, title: str | None, body: str, *, border_style: str = "yellow"
    ) -> None:
        self.printed.append(("panel", title, body, border_style))

    def step(self, title: str, *, number: int) -> None:
        self.steps.append((title, number))

    def print(self, text: object = "", *, markup: bool | None = None) -> None:
        _ = markup
        self.printed.append(text)

    async def confirm(self, _prompt: str, default: bool = True) -> bool | None:
        _ = default
        return next(self.confirms)

    async def select(self, _prompt: str, choices: list[tuple[str, Any]]) -> Any | None:
        _ = choices
        return next(self.selects)

    async def password(self, _prompt: str) -> str | None:
        return next(self.passwords)


class DummyServices:
    def __init__(
        self,
        *,
        bot_info: list[User | None] | None = None,
        chat: onboarding.ChatInfo | None = None,
        topics_issue: ConfigError | None = None,
        engines: list[tuple[str, bool, str | None]] | None = None,
        config: dict[str, Any] | None = None,
    ) -> None:
        self.bot_info = iter(bot_info or [])
        self.chat = chat
        self.topics_issue = topics_issue
        self.engines = engines or []
        self.config = config or {}
        self.writes: list[tuple[Path, dict[str, Any]]] = []

    async def get_bot_info(self, _token: str) -> User | None:
        return next(self.bot_info)

    async def wait_for_chat(self, _token: str) -> onboarding.ChatInfo:
        assert self.chat is not None
        return self.chat

    async def validate_topics(
        self, _token: str, _chat_id: int, _scope: onboarding.TopicScope
    ) -> ConfigError | None:
        return self.topics_issue

    def list_engines(self) -> list[tuple[str, bool, str | None]]:
        return list(self.engines)

    def read_config(self, _path: Path) -> dict[str, Any]:
        return dict(self.config)

    def write_config(self, path: Path, data: dict[str, Any]) -> None:
        self.writes.append((path, data))


def test_chat_info_display_and_kind() -> None:
    chat = onboarding.ChatInfo(
        chat_id=1,
        username="alice",
        title=None,
        first_name=None,
        last_name=None,
        chat_type="private",
    )
    assert chat.display == "@alice"
    assert chat.kind == "private chat"

    group = onboarding.ChatInfo(
        chat_id=2,
        username=None,
        title="Team",
        first_name=None,
        last_name=None,
        chat_type="supergroup",
    )
    assert group.display == 'group "Team"'
    assert group.kind == 'supergroup "Team"'

    channel = onboarding.ChatInfo(
        chat_id=3,
        username=None,
        title=None,
        first_name=None,
        last_name=None,
        chat_type="channel",
    )
    assert channel.display == "channel"
    assert channel.kind == "channel"

    unnamed = onboarding.ChatInfo(
        chat_id=4,
        username=None,
        title=None,
        first_name="Ada",
        last_name="Lovelace",
        chat_type=None,
    )
    assert unnamed.display == "Ada Lovelace"
    assert unnamed.kind == "private chat"


def test_onboarding_state_helpers(tmp_path: Path) -> None:
    state = onboarding.OnboardingState(config_path=tmp_path / "cfg", force=False)
    assert state.is_stateful is False
    assert state.bot_ref == "your bot"

    state.session_mode = "chat"
    assert state.is_stateful is True

    state.session_mode = "stateless"
    state.topics_enabled = True
    assert state.is_stateful is True

    state.bot_name = "Takopi"
    assert state.bot_ref == "Takopi"
    state.bot_username = "yee88_bot"
    assert state.bot_ref == "@yee88_bot"


def test_display_path(tmp_path: Path) -> None:
    home_path = Path.home() / "yee88" / "cfg.toml"
    assert onboarding.display_path(home_path).startswith("~/")
    assert onboarding.display_path(tmp_path / "cfg.toml") == str(tmp_path / "cfg.toml")


def test_build_transport_patch_requires_fields(tmp_path: Path) -> None:
    state = onboarding.OnboardingState(config_path=tmp_path / "cfg", force=False)
    with pytest.raises(RuntimeError, match="missing chat"):
        onboarding.build_transport_patch(state, bot_token="x")

    state.chat = onboarding.ChatInfo(
        chat_id=1,
        username=None,
        title=None,
        first_name=None,
        last_name=None,
        chat_type="private",
    )
    with pytest.raises(RuntimeError, match="missing session mode"):
        onboarding.build_transport_patch(state, bot_token="x")

    state.session_mode = "chat"
    with pytest.raises(RuntimeError, match="missing resume choice"):
        onboarding.build_transport_patch(state, bot_token="x")


def test_build_config_patch_and_merge(tmp_path: Path) -> None:
    state = onboarding.OnboardingState(config_path=tmp_path / "cfg", force=False)
    state.chat = onboarding.ChatInfo(
        chat_id=10,
        username=None,
        title=None,
        first_name=None,
        last_name=None,
        chat_type="private",
    )
    state.session_mode = "chat"
    state.show_resume_line = False
    state.default_engine = "codex"
    state.topics_enabled = True
    state.topics_scope = "all"

    patch = onboarding.build_config_patch(state, bot_token="token")
    assert patch["default_engine"] == "codex"

    merged = onboarding.merge_config(
        {"bot_token": "old", "chat_id": 1, "transports": {}},
        patch,
        config_path=tmp_path / "cfg.toml",
    )
    assert merged["transport"] == "telegram"
    assert merged["transports"]["telegram"]["bot_token"] == "token"
    assert merged["transports"]["telegram"]["topics"]["enabled"] is True
    assert "bot_token" not in merged
    assert "chat_id" not in merged


def test_render_helpers() -> None:
    text = onboarding.render_botfather_instructions()
    assert "BotFather" in text.plain
    assert "send /newbot" in text.plain

    topics = onboarding.render_topics_group_instructions("@bot")
    assert "topics" in topics.plain

    generic = onboarding.render_generic_capture_prompt("@bot")
    assert "send /start" in generic.plain

    warning = onboarding.render_topics_validation_warning(ConfigError("boom"))
    assert "warning" in warning.plain

    config_warning = onboarding.render_config_malformed_warning(ConfigError("bad"))
    assert "config is malformed" in config_warning.plain

    backup_warning = onboarding.render_backup_failed_warning(OSError("nope"))
    assert "failed to back up" in backup_warning.plain

    tabs = onboarding.render_persona_tabs()
    assert isinstance(tabs, Table)

    preview = onboarding.render_workspace_preview()
    assert "memory-box" in preview.plain

    assistant = onboarding.render_assistant_preview()
    assert "make happy wings fit" in assistant.plain

    handoff = onboarding.render_handoff_preview()
    assert "resume" in handoff.plain

    convo = Text()
    onboarding.append_dialogue(convo, "bot", "hello", speaker_style="cyan")
    assert "hello" in convo.plain


def test_render_engine_table_prints() -> None:
    ui = DummyUI()
    onboarding.render_engine_table(
        cast(onboarding.UI, ui),
        [("codex", True, None), ("other", False, "brew install other")],
    )
    assert ui.printed


def test_debug_onboarding_paths_prints_table() -> None:
    console = Console(record=True, width=120)
    onboarding.debug_onboarding_paths(console=console)
    output = console.export_text()
    assert "onboarding paths (15)" in output
    assert "workspace" in output
    assert "assistant" in output
    assert "handoff" in output


@pytest.mark.anyio
async def test_confirm_prompt_returns_question_result(monkeypatch) -> None:
    seen: dict[str, object] = {}

    class _PromptSession:
        def __init__(self, *args, **kwargs) -> None:
            seen["args"] = args
            seen["kwargs"] = kwargs
            self.app = object()

    class _Question:
        def __init__(self, app) -> None:
            seen["app"] = app

        async def ask_async(self) -> bool | None:
            return True

    monkeypatch.setattr(onboarding, "PromptSession", _PromptSession)
    monkeypatch.setattr(onboarding, "Question", _Question)

    result = await onboarding.confirm_prompt("continue?", default=False)

    assert result is True
    assert "kwargs" in seen


@pytest.mark.anyio
async def test_get_bot_info_retries(monkeypatch) -> None:
    class _Bot:
        def __init__(self, _token: str) -> None:
            self.calls = 0
            self.closed = False

        async def get_me(self) -> User | None:
            self.calls += 1
            if self.calls < 3:
                raise TelegramRetryAfter(0)
            return User(id=1, username="bot")

        async def close(self) -> None:
            self.closed = True

    async def _sleep(_seconds: float) -> None:
        return None

    monkeypatch.setattr(onboarding, "TelegramClient", _Bot)
    monkeypatch.setattr(onboarding.anyio, "sleep", _sleep)

    info = await onboarding.get_bot_info("token")
    assert info is not None
    assert info.username == "bot"


@pytest.mark.anyio
async def test_get_bot_info_gives_up(monkeypatch) -> None:
    class _Bot:
        def __init__(self, _token: str) -> None:
            self.calls = 0

        async def get_me(self) -> User | None:
            self.calls += 1
            raise TelegramRetryAfter(0)

        async def close(self) -> None:
            return None

    async def _sleep(_seconds: float) -> None:
        return None

    monkeypatch.setattr(onboarding, "TelegramClient", _Bot)
    monkeypatch.setattr(onboarding.anyio, "sleep", _sleep)

    info = await onboarding.get_bot_info("token")
    assert info is None


@pytest.mark.anyio
async def test_wait_for_chat_filters_updates(monkeypatch) -> None:
    updates = [
        [
            Update(
                update_id=1,
                message=Message(
                    message_id=1,
                    from_=User(id=1, is_bot=True),
                    chat=Chat(id=1, type="private"),
                ),
            )
        ],
        None,
        [],
        [Update(update_id=2, message=None)],
        [
            Update(
                update_id=3,
                message=Message(
                    message_id=3,
                    from_=User(id=2, is_bot=True),
                    chat=Chat(id=2, type="private"),
                ),
            )
        ],
        [
            Update(
                update_id=4,
                message=Message(
                    message_id=4,
                    from_=User(id=3, is_bot=True),
                    chat=Chat(id=3, type="private"),
                ),
            )
        ],
        [
            Update(
                update_id=5,
                message=Message(
                    message_id=5,
                    from_=User(id=4, is_bot=True),
                    chat=Chat(id=4, type="private"),
                ),
            )
        ],
        [
            Update(
                update_id=6,
                message=Message(
                    message_id=6,
                    from_=User(id=5, is_bot=False),
                    chat=Chat(id=7, username="bob", type="private"),
                ),
            )
        ],
    ]

    class _Bot:
        def __init__(self, _token: str) -> None:
            self.calls = 0

        async def get_updates(self, *args, **kwargs):
            _ = args, kwargs
            idx = self.calls
            self.calls += 1
            return updates[idx]

        async def close(self) -> None:
            return None

    async def _sleep(_seconds: float) -> None:
        return None

    monkeypatch.setattr(onboarding, "TelegramClient", _Bot)
    monkeypatch.setattr(onboarding.anyio, "sleep", _sleep)

    chat = await onboarding.wait_for_chat("token")
    assert chat.chat_id == 7
    assert chat.username == "bob"


@pytest.mark.anyio
async def test_validate_topics_onboarding_errors(monkeypatch) -> None:
    class _Bot:
        def __init__(self, _token: str) -> None:
            return None

        async def close(self) -> None:
            return None

    async def _raise_config(*_args, **_kwargs):
        raise ConfigError("bad")

    async def _raise_other(*_args, **_kwargs):
        raise RuntimeError("boom")

    monkeypatch.setattr(onboarding, "TelegramClient", _Bot)
    monkeypatch.setattr(onboarding, "_validate_topics_setup_for", _raise_config)
    err = await onboarding.validate_topics_onboarding("token", 1, "auto", ())
    assert isinstance(err, ConfigError)

    monkeypatch.setattr(onboarding, "_validate_topics_setup_for", _raise_other)
    err = await onboarding.validate_topics_onboarding("token", 1, "auto", ())
    assert isinstance(err, ConfigError)
    assert "topics validation failed" in str(err)


@pytest.mark.anyio
async def test_prompt_token_success_and_retry() -> None:
    ui = DummyUI(confirms=[True], passwords=["", "token", "token"])  # empty then retry
    svc = DummyServices(
        bot_info=[None, User(id=1, username="bot")],
    )

    token, info = await onboarding.prompt_token(
        cast(onboarding.UI, ui), cast(onboarding.Services, svc)
    )
    assert token == "token"
    assert info.username == "bot"


@pytest.mark.anyio
async def test_prompt_token_cancel_on_failure() -> None:
    ui = DummyUI(confirms=[False], passwords=["token"])
    svc = DummyServices(bot_info=[None])

    with pytest.raises(onboarding.OnboardingCancelled):
        await onboarding.prompt_token(
            cast(onboarding.UI, ui), cast(onboarding.Services, svc)
        )


@pytest.mark.anyio
async def test_capture_chat_sets_state() -> None:
    chat = onboarding.ChatInfo(
        chat_id=5,
        username=None,
        title="Team",
        first_name=None,
        last_name=None,
        chat_type="group",
    )
    ui = DummyUI()
    svc = DummyServices(chat=chat)
    state = onboarding.OnboardingState(config_path=Path("/tmp/cfg"), force=False)
    state.token = "token"

    await onboarding.capture_chat(
        cast(onboarding.UI, ui), cast(onboarding.Services, svc), state
    )
    assert state.chat == chat


@pytest.mark.anyio
async def test_step_capture_chat_workspace_switches_to_assistant(
    tmp_path: Path,
) -> None:
    chat = onboarding.ChatInfo(
        chat_id=6,
        username=None,
        title=None,
        first_name="Ada",
        last_name=None,
        chat_type="private",
    )
    ui = DummyUI(selects=["assistant"])
    svc = DummyServices(chat=chat, topics_issue=ConfigError("nope"))
    state = onboarding.OnboardingState(config_path=tmp_path / "cfg", force=False)
    state.token = "token"
    state.bot_name = "Takopi"
    state.persona = "workspace"
    state.topics_scope = "auto"
    state.topics_enabled = True

    await onboarding.step_capture_chat(
        cast(onboarding.UI, ui), cast(onboarding.Services, svc), state
    )

    assert state.persona == "assistant"
    assert state.topics_enabled is False


@pytest.mark.anyio
async def test_step_default_engine_no_installed(tmp_path: Path) -> None:
    ui = DummyUI(confirms=[False])
    svc = DummyServices(engines=[("codex", False, None)])
    state = onboarding.OnboardingState(config_path=tmp_path / "cfg", force=False)

    with pytest.raises(onboarding.OnboardingCancelled):
        await onboarding.step_default_engine(
            cast(onboarding.UI, ui), cast(onboarding.Services, svc), state
        )



from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import Any

from ..model import ActionKind
from ..utils.paths import relativize_command, relativize_path


def tool_input_path(
    tool_input: Mapping[str, Any],
    *,
    path_keys: Sequence[str],
) -> str | None:
    for key in path_keys:
        value = tool_input.get(key)
        if isinstance(value, str) and value:
            return value
    return None


def tool_kind_and_title(
    tool_name: str,
    tool_input: Mapping[str, Any],
    *,
    path_keys: Sequence[str],
    task_kind: ActionKind = "subagent",
) -> tuple[ActionKind, str]:
    name_lower = tool_name.lower()

    if name_lower in {"bash", "shell", "killshell"}:
        command = tool_input.get("command")
        display = relativize_command(str(command or tool_name))
        return "command", display

    if name_lower in {"edit", "write", "notebookedit", "multiedit"}:
        path = tool_input_path(tool_input, path_keys=path_keys)
        if path:
            return "file_change", relativize_path(str(path))
        return "file_change", str(tool_name)

    if name_lower == "read":
        path = tool_input_path(tool_input, path_keys=path_keys)
        if path:
            return "tool", f"read: `{relativize_path(str(path))}`"
        return "tool", "read"

    if name_lower == "glob":
        pattern = tool_input.get("pattern")
        if pattern:
            return "tool", f"glob: `{pattern}`"
        return "tool", "glob"

    if name_lower == "grep":
        pattern = tool_input.get("pattern")
        if pattern:
            return "tool", f"grep: {pattern}"
        return "tool", "grep"

    if name_lower == "find":
        pattern = tool_input.get("pattern")
        if pattern:
            return "tool", f"find: {pattern}"
        return "tool", "find"

    if name_lower == "ls":
        path = tool_input_path(tool_input, path_keys=path_keys)
        if path:
            return "tool", f"ls: `{relativize_path(str(path))}`"
        return "tool", "ls"

    if name_lower in {"websearch", "web_search"}:
        query = tool_input.get("query")
        return "web_search", str(query or "search")

    if name_lower in {"webfetch", "web_fetch"}:
        url = tool_input.get("url")
        return "web_search", str(url or "fetch")

    if name_lower in {"todowrite", "todoread"}:
        return "note", "update todos" if "write" in name_lower else "read todos"

    if name_lower == "askuserquestion":
        return "note", "ask user"

    if name_lower in {"task", "agent"}:
        desc = tool_input.get("description") or tool_input.get("prompt")
        return task_kind, str(desc or tool_name)

    return "tool", tool_name
